# Copyright 2024 David Trimm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import heapq

import networkx as nx
import numpy as np

import squarify # treemaps
import random

from math import sqrt, pi, cos, sin, ceil, floor, inf, atan2
from dataclasses import dataclass

from multiprocessing import Pool

__name__ = 'rt_graph_layouts_mixin'

#
# Graph Layouts Methods
#
class RTGraphLayoutsMixin(object):
    #
    # positionExtents()
    # - extents of all the nodes in the positions dictionary
    # - will add x or y space if there's only a single x or y coordinate
    #
    def positionExtents(self,
                        pos,            # pos['node'] = [x, y]
                        _graph = None): # if specified, will limit calculation to just the nodes in this graph
        x0 = y0 = x1 = y1 = None

        if _graph is not None:
            from_structure = set(_graph.nodes())
            in_pos         = set(pos.keys())
            if len(in_pos & from_structure) != len(from_structure):
                print(f'{in_pos=}\n{from_structure=}')
                raise Exception(f'Missing keys in position dictionary | AND={len(in_pos & from_structure)} SUBG={len(from_structure)}')
        else:  from_structure = set(pos.keys())

        for _node in from_structure:
            x = pos[_node][0]
            y = pos[_node][1]
            x0 = x if x0 is None else min(x0, x)
            y0 = y if y0 is None else min(y0, y)
            x1 = x if x1 is None else max(x1, x)
            y1 = y if y1 is None else max(y1, y)
        if x0 == x1:
            x0, x1 = x0 - 0.5, x1 + 0.5
        if y0 == y1:
            y0, y1 = y0 - 0.5, y1 + 0.5
        return x0,y0,x1,y1

    #
    # calculateLevelSet()
    #
    def calculateLevelSet(self,
                          pos,                   # dictionary of nodes to array of positions -- needs to be two (or more)
                          bounds_percent = .05,  # inset the graph into the view by this percent... so that the nodes aren't right at the edges 
                          w              = 256,  # view width to use for level set
                          h              = 256): # view height to use for level set
        # Determine the min and max positions
        x0,y0,x1,y1 = self.positionExtents(pos)

        # Provide border space
        x_inc = (x1-x0)*bounds_percent
        x0 -= x_inc
        x1 += x_inc
        y_inc = (y1-y0)*bounds_percent
        y0 -= y_inc
        y1 += y_inc

        # Translation lambdas
        xT = lambda x: int(w*(x-x0)/(x1-x0))
        yT = lambda y: int(h*(y-y0)/(y1-y0))

        # Allocate the level set
        node_info  = [[None for x in range(w)] for y in range(h)] # node that found the pixel
        found_time = [[None for x in range(w)] for y in range(h)] # when node was found

        # Distance lambda function
        dist = lambda _x0,_y0,_x1,_y1: sqrt((_x0-_x1)*(_x0-_x1)+(_y0-_y1)*(_y0-_y1))

        # Initialize the level set with the node positions
        _node_pos = {}
        _heap     = []
        for _node in pos.keys():
            xi = xT(pos[_node][0])
            yi = yT(pos[_node][1])
            _node_pos[_node] = (xi,yi)
            node_info [yi][xi] = _node
            found_time[yi][xi] = 0
            for dx in range(-1,2):
                for dy in range(-1,2):
                    xn = xi + dx
                    yn = yi + dy
                    if (dx == 0 and dy == 0) or xn < 0  or yn < 0 or xn >= w or yn >= h:
                        continue
                    else:
                        t  = dist(xi,yi,xn,yn)
                        heapq.heappush(_heap,(t, xn, yn, _node))

        # Go through the heap
        while len(_heap) > 0:
            t,xi,yi,_node = heapq.heappop(_heap)
            if found_time[yi][xi] is None or found_time[yi][xi] > t:
                node_info [yi][xi] = _node
                found_time[yi][xi] = t
                for dx in range(-1,2):
                    for dy in range(-1,2):
                        xn = xi + dx
                        yn = yi + dy
                        if (dx == 0 and dy == 0) or xn < 0  or yn < 0 or xn >= w or yn >= h:
                            continue
                        else:
                            xp,yp = _node_pos[_node]
                            t  = dist(xp,yp,xn,yn)
                            if found_time[yn][xn] is None or \
                               found_time[yn][xn] > t:
                                heapq.heappush(_heap,(t, xn, yn, _node))

        return node_info,found_time


    #
    # levelSetSVG()
    # - create a level set representation in svg format
    #
    def levelSetSVG(self,
                    node_info,
                    found_time):
        # Determine width and height of the levelset
        h = len(node_info)
        w = len(node_info[0])

        # Determine the maximum value
        max_t = 1
        for yi in range(0,h):
            for xi in range(0,w):
                if found_time[yi][xi] is not None and found_time[yi][xi] > max_t:
                    max_t = found_time[yi][xi]

        svg  = f'<svg x="0" y="0" width="{2*w}" height="{h}">'
        for yi in range(0,h):
            for xi in range(0,w):
                if found_time[yi][xi] is not None:
                    _co = self.co_mgr.spectrum(found_time[yi][xi],0,max_t)
                    svg += f'<rect x="{xi}" y="{yi}" width="1" height="1" fill="{_co}" stroke="none" stroke-opacity="0.0" />'
                    _co = self.co_mgr.getColor(node_info[yi][xi])
                    svg += f'<rect x="{xi+w}" y="{yi}" width="1" height="1" fill="{_co}" stroke="none" stroke-opacity="0.0" />'

        svg += '</svg>'
        return svg
        
    #
    # adjustNodePositionsBasedOnLevelSet()
    # ... doesn't produce good results... does fill in the whitespace but without regards to edges
    #
    def adjustNodePositionsBasedOnLevelSet(self,
                                           node_info,
                                           pos):
        # Determine width and height of the levelset
        h = len(node_info)
        w = len(node_info[0])

        # Find the center of mass and place the node there
        xsum    = {}
        ysum    = {}
        samples = {}
        for yi in range(0,h):
            for xi in range(0,w):
                if node_info[yi][xi] is not None:
                    _node = node_info[yi][xi]
                    if _node not in xsum.keys():
                        xsum[_node] = 0
                        ysum[_node] = 0
                        samples[_node] = 0
                    xsum[_node] += xi
                    ysum[_node] += yi
                    samples[_node] += 1

        new_pos = {}
        for _node in samples.keys():
            new_pos[_node] = [xsum[_node]/samples[_node], ysum[_node]/samples[_node]]

        not_set = set(pos.keys()) - set(new_pos.keys())
        for x in not_set:
            new_pos[x] = pos[x]

        return new_pos

    #
    # __highDegreeNodes__() - returns higher degree nodes.. could've done better...
    #
    def __highDegreeNodes__(self, _graph):
        _nodes   = _graph.nodes()
        _degrees = []
        for _node in _nodes: _degrees.append(_graph.degree(_node))
        _degrees.sort()
        _degrees.reverse()
        top_ten = _degrees[:int(len(_degrees)*0.1)]
        return top_ten

    #
    # rectangularArrangement() - arrange a list of nodes in a rectangular shape.
    # - bounds = (x0,y0,x1,y1) where x0 < x1 and y0 < y1
    #
    def rectangularArrangement(self, g, nodes, pos=None, bounds=(0,0,1,1)):
        x0, y0, x1, y1 = bounds
        if x0 > x1: x0, x1 = x1, x0
        if y0 > y1: y0, y1 = y1, y0
        dx, dy = float(x1-x0), float(y1-y0)
        if type(nodes) is not list: nodes = list(nodes)
        if pos is None: pos = {}
        if   len(nodes) == 1:
            pos[nodes[0]] = (x0 + dx/2.0, y0 + dy/2.0)
        elif len(nodes) == 2:
            pos[nodes[0]], pos[nodes[1]] = (x0, y0+dy/2), (x1, y0+dy/2.0)
        elif len(nodes) == 3:
            pos[nodes[0]], pos[nodes[1]], pos[nodes[2]] = (x0, y0), (x0 + dx/2.0, y1), (x1, y0)
        elif len(nodes) == 4 or len(nodes) == 5:
            pos[nodes[0]], pos[nodes[1]] = (x0,y0),(x1,y0)
            pos[nodes[2]], pos[nodes[3]] = (x0,y1),(x1,y1)
            if len(nodes) == 5: pos[nodes[4]] = ((x0+x1)/2.0, (y0+y1)/2.0)
        else:
            if x0 >= x1: x1 = x0 + 1.0
            if y0 >= y1: y1 = y0 + 1.0
            dx, dy = x1 - x0, y1 - y0
            n = ceil(sqrt(len(nodes)))
            if (dx/dy) > 1.5 or (dy/dx) > 1.5: # rectangular
                closest_d = inf
                for i in range(1,n+1):
                    other = len(nodes)/i
                    ratio = other/i
                    d     = abs(ratio - dx/dy)
                    if d < closest_d:
                        max_x_i = i     if (i > other) else other
                        max_y_i = other if (i > other) else i
                    else:
                        max_x_i = other if (i > other) else i
                        max_y_i = i     if (i > other) else other
            else:                              # roughly square
                max_x_i = max_y_i = n

            _sorter_  = []
            _degrees_ = g.degree(nodes)
            for node in nodes: 
                _degrees_ = g.degree(node)
                if type(_degrees_) == int: _sorter_.append((_degrees_,      node))
                else:                      _sorter_.append((len(_degrees_), node))
            _sorter_ = sorted(_sorter_, reverse=True)

            x_i, y_i = 0, 0
            for i in range(len(nodes)):
                _x_ = x0 + x_i * (dx/max_x_i)
                _y_ = y0 + y_i * (dy/max_y_i)
                pos[_sorter_[i][1]] = (_x_,_y_)
                x_i += 1
                if x_i >= max_x_i:
                    y_i += 1
                    x_i  = 0

        return pos

    #
    # sunflowerSeedArrangement() - arrange a list of nodes in a sunflower arrangement
    #
    def sunflowerSeedArrangement(self, g, nodes, pos=None, xy=None, r_max=1.0):
        if type(nodes) is not list: nodes = list(nodes)
        if xy is None: xy = (0,0)
        n = len(nodes)

        # place highest degree nodes in the center
        _sorter_  = []
        _degrees_ = g.degree(nodes)
        for node in nodes: 
            _degrees_ = g.degree(node)
            if type(_degrees_) == int: _sorter_.append((_degrees_,      node))
            else:                      _sorter_.append((len(_degrees_), node))
        _sorter_ = sorted(_sorter_, reverse=True)

        if pos is None:  pos = {}
        r_max_formula = np.sqrt(n)
        _golden_ratio_ = (1 + np.sqrt(5)) / 2
        for i in range(n):
            _angle_  = i * 2 * np.pi / _golden_ratio_
            _radius_ = r_max * np.sqrt(i) / r_max_formula
            pos[_sorter_[i][1]] = (xy[0] + _radius_ * np.cos(_angle_), 
                                   xy[1] + _radius_ * np.sin(_angle_))
        return pos


    #
    # linearOptimizedArrangement() - attempt to place the nodes in their best positions in a line
    # - _segment_ = [(x0,y0),(x1,y1)]
    #
    def linearOptimizedArrangement(self, g, nodes, pos, segment=((0.0, 0.0), (1.0, 1.0))):
        if len(nodes) == 1: return {nodes[0]:((segment[0][0]+segment[1][0])/2.0, (segment[0][1]+segment[1][1])/2.0)}
        adj_pos, as_set = {}, set(nodes)
        # Break the nodes into externally (any) connected and internally (only) connected
        _externals_, _internals_ = set(), set()
        for _node_ in nodes:
            for _nbor_ in g.neighbors(_node_):
                if _nbor_ not in as_set: 
                    _externals_.add(_node_)  # any external connection, makes it external
                    break
            if _node_ not in _externals_: _internals_.add(_node_) # if it wasn't added in the loop, it's an internal

        # Create blanks for positions on the line
        dx, dy = segment[1][0] - segment[0][0], segment[1][1] - segment[0][1]
        _filled_with_, _locations_ = [], []
        for i in range(len(nodes)): 
            _filled_with_.append(None)
            perc = i/float(len(nodes)-1)
            _locations_.append((segment[0][0] + dx*perc, segment[0][1] + dy*perc))

        def placeNodeIntoClosestSlot(node_to_place, nodes_xy=None):
            _closest_ = 0
            if nodes_xy is not None:
                _closest_pt_ = self.closestPointOnSegment(segment, nodes_xy)[1]
                if    _closest_pt_ == segment[0]: _closest_ = 0
                elif  _closest_pt_ == segment[1]: _closest_ = len(_locations_)-1
                else:
                    _closest_d_ = sqrt((_closest_pt_[0]-_locations_[0][0])**2 + (_closest_pt_[1]-_locations_[0][1])**2)
                    for i in range(1, len(_locations_)):
                        _d_ = sqrt((_closest_pt_[0]-_locations_[i][0])**2 + (_closest_pt_[1]-_locations_[i][1])**2)
                        if _d_ < _closest_d_: _closest_, _closest_d_ = i, _d_

            if _filled_with_[_closest_] is None: 
                _filled_with_[_closest_] = node_to_place
                adj_pos[node_to_place]   = _locations_[_closest_]
            else:
                for j in range(1, len(_locations_)):
                    up = _closest_+j
                    dn = _closest_-j
                    if   up < len(_locations_) and _filled_with_[up] is None:
                        _filled_with_[up]        = node_to_place
                        adj_pos[node_to_place]   = _locations_[up]
                        break
                    elif dn >= 0 and _filled_with_[dn] is None:
                        _filled_with_[dn]        = node_to_place
                        adj_pos[node_to_place]   = _locations_[dn]
                        break

        # Place the external nodes -- start with the highest degree node
        _sorter_ = []
        for _node_ in _externals_: _sorter_.append((g.degree(_node_), _node_))
        _sorter_ = sorted(_sorter_, reverse=True)
        for _tuple_ in _sorter_:
            _node_ = _tuple_[1]
            _x_sum_, _y_sum_, _samples_ = 0.0, 0.0, 0
            for _nbor_ in g.neighbors(_node_):
                if _nbor_ in as_set: continue
                _x_sum_   += pos[_nbor_][0]
                _y_sum_   += pos[_nbor_][1]
                _samples_ += 1
            _x_, _y_ = _x_sum_ / _samples_, _y_sum_ / _samples_
            placeNodeIntoClosestSlot(_node_, (_x_, _y_))
        
        # Place the rest of the nodes
        for _node_ in _internals_: placeNodeIntoClosestSlot(_node_)

        return adj_pos

    #
    # circularOptimizedArrangement() - attempt to place the nodes in their best positions around a circle
    # - returns a dictionary of the nodes that were adjusted -- should be equal to the nodes passed in
    #
    def circularOptimizedArrangement(self, g, nodes, pos, xy=(0.0,0.0), r=1.0):
        adj_pos, as_set  = {}, set(nodes)

        # Break the nodes into externally (any) connected and internally (only) connected
        _externals_, _internals_ = set(), set()
        for _node_ in nodes:
            for _nbor_ in g.neighbors(_node_):
                if _nbor_ not in as_set: 
                    _externals_.add(_node_)  # any external connection, makes it external
                    break
            if _node_ not in _externals_: _internals_.add(_node_) # if it wasn't added in the loop, it's an internal

        # Create blanks for positions around the circle
        _filled_with_, _angulars_, _angular_locations_ = [], [], []
        for i in range(len(nodes)):
            _angle_            = i * 2 * pi / len(nodes)
            _filled_with_      .append(None)
            _angulars_         .append(_angle_)
            _angular_locations_.append((xy[0] + r * cos(_angle_), xy[1] + r * sin(_angle_)))

        # For a given angle, find the closest available slot to place the node
        def placeNodeIntoClosestSlot(node_to_place, nodes_xy=None):
            _closest_ = 0
            if nodes_xy is None: _closest_ = random.randint(0, len(_angulars_)-1)
            else:
                _closest_, _closest_d_ = 0, 1e9
                for i in range(0, len(_angular_locations_)):
                    dx, dy = nodes_xy[0] - _angular_locations_[i][0], nodes_xy[1] - _angular_locations_[i][1]
                    d      = sqrt(dx*dx + dy*dy)
                    if d < _closest_d_: _closest_, _closest_d_ = i, d
                    
            if _filled_with_[_closest_] is None: 
                _filled_with_[_closest_] = node_to_place
                adj_pos[node_to_place]   = _angular_locations_[_closest_]
            else:
                for j in range(1, len(_angulars_)):
                    up =  (_closest_+j)                  % len(_angulars_)
                    dn = ((_closest_-j)+len(_angulars_)) % len(_angulars_)
                    if _filled_with_[up] is None:
                        _filled_with_[up]        = node_to_place
                        adj_pos[node_to_place]   = _angular_locations_[up]
                        break
                    elif _filled_with_[dn] is None:
                        _filled_with_[dn]        = node_to_place
                        adj_pos[node_to_place]   = _angular_locations_[dn]
                        break

        # Arrange the externally connected nodes first ... put them in the closest slot
        for _node_ in _externals_:
            _x_sum_, _y_sum_, _samples_ = 0.0, 0.0, 0
            for _nbor_ in g.neighbors(_node_):
                if _nbor_ not in _internals_ and _nbor_ not in _externals_:
                    _nbor_xy_   =  pos[_nbor_]
                    _x_sum_     += _nbor_xy_[0]
                    _y_sum_     += _nbor_xy_[1]
                    _samples_   += 1
            _x_, _y_ = _x_sum_ / _samples_, _y_sum_ / _samples_
            placeNodeIntoClosestSlot(_node_, (_x_,_y_))

        # Arrange the internally connected nodes next ... start with the ones connected to the most externals
        _sorter_ = []
        for _node_ in _internals_:
            _externally_connected_ = 0
            for _nbor_ in g.neighbors(_node_):
                if _nbor_ in _externals_: _externally_connected_ += 1
            _sorter_.append((_externally_connected_, _node_))
        _sorter_ = sorted(_sorter_, reverse=True)
        for _tuple_ in _sorter_:
            _node_ = _tuple_[1]
            if _tuple_[0] > 0:
                _x_sum_, _y_sum_, _samples_ = 0.0, 0.0, 0
                for _nbor_ in g.neighbors(_node_):
                    if _nbor_ in _filled_with_:
                        i           =  _filled_with_.index(_nbor_)
                        _x_sum_     += _angular_locations_[i][0]
                        _y_sum_     += _angular_locations_[i][1]
                        _samples_   += 1
                _x_, _y_ = _x_sum_ / _samples_, _y_sum_ / _samples_
                placeNodeIntoClosestSlot(_node_, (_x_,_y_))
            else:
                placeNodeIntoClosestSlot(_node_)

        return adj_pos

    #
    # circularLayout() - from the java version
    #
    def circularLayout(self, g, selection=None, _radius_=100):
        if selection is None: selection = self.__highDegreeNodes__(g)

        pos, center_angle = {}, {}
        for i in range(len(selection)):
            _node = selection[i]
            _angle_ = i*2*pi/len(selection)
            center_angle[_node] = _angle_
            pos[_node] = (_radius_*cos(_angle_),_radius_*sin(_angle_))

        outer_rings   = {}
        _plus_        = _radius_ * 0.2
        _radius_plus_ = _radius_ + _plus_ + _plus_
        for _node in g.nodes():
            if _node not in pos.keys():
                attachments        = set()
                attachments_coords = set()
                for _center in selection:
                    if _center in g[_node]:
                        attachments.add(_center)
                        attachments_coords.add(pos[_center])
                if   len(attachments) == 0:
                    pos[_node] = (0,0)
                elif len(attachments) == 1:
                    center      = attachments.pop()
                    if center not in outer_rings: outer_rings[center] = set()
                    outer_rings[center].add(_node)
                else:
                    x_sum, y_sum = 0, 0
                    for xy in attachments_coords:
                        x_sum += xy[0]
                        y_sum += xy[1]
                    _subangle_  = 2*pi*random.random()
                    _subradius_ = _plus_ * random.random()
                    pos[_node] = (x_sum/len(attachments_coords) + _subradius_*cos(_subangle_), 
                                  y_sum/len(attachments_coords) + _subradius_*sin(_subangle_))

        # Layout the outer rings using a sunflower arrangement
        for center in outer_rings.keys():
            _angle_  = center_angle[center]
            xy       = (_radius_plus_*cos(_angle_), _radius_plus_*sin(_angle_))
            pos      = self.sunflowerSeedArrangement(g, outer_rings[center], pos, xy, _plus_)

        return pos
    
    #
    # Count the nodes in a subtree of a tree
    #
    def __countSubTreeNodes__(self, _graph, _node, _ignore, _child_count):
        # Check the cache
        if _node in _child_count.keys():
            return _child_count[_node] + 1 # children plus this node
        
        # Else recursively count children
        _sum = 0
        for x in _graph[_node]:
            if x == _ignore:
                continue
            _sum += self.__countSubTreeNodes__(_graph, x, _node, _child_count)

        # Cache the value
        if _child_count is not None:
            _child_count[_node] = _sum
        
        # Results are this node plus children
        return _sum+1

    #
    # Count the total number of leaves
    # ... recursive
    #
    def __totalLeaves__(self, _graph, _parent, _node, _leaf_count):
        if len(_graph[_node]) == 1:
            _leaf_count[_node] = 0 # it's a leaf...
            return 1
        else:
            _sum = 0
            for x in _graph[_node]:
                if x != _parent:
                    _sum += self.__totalLeaves__(_graph, _node, x, _leaf_count)
            _leaf_count[_node] = _sum
            return _sum

    #
    # Calculate the depth of the tree
    # ... recursive
    #
    def __treeDepth__(self, _graph, _parent, _node):
        if len(_graph[_node]) == 1:
            return 1
        else:
            _max_depth = 0
            for x in _graph[_node]:
                if x != _parent:
                    _depth = self.__treeDepth__(_graph, _node, x)
                    if _depth > _max_depth:
                        _max_depth = _depth
            return _max_depth + 1

    #
    # dagLeavesOnly() - for a directed acycle graph, return a set of the leaves
    #
    def dagLeavesOnly(self, G):
        leaves = set()
        for node in G.nodes():
            nbor_count = 0
            for nbor in G.neighbors(node):
                nbor_count += 1
            if nbor_count == 1:
                leaves.add(node)
        return leaves

    #
    # hyperTreeLayout()
    # - create a hypertree layout
    #
    def hyperTreeLayout(self,
                        _graph,                        # networkx graph
                        roots                 = None,  # root(s) to use... if not set, will be calculated
                        touch_up_with_springs = False, # touch up the center of the layout with a spring layout
                        bounds_percent        = 0.1):  # for tree map positioning
        # Make sure root is a list
        if type(roots) != list: roots = [roots]

        # Separate graph into connected components
        _graph = nx.to_undirected(_graph)
        S = [_graph.subgraph(c).copy() for c in nx.connected_components(_graph)]

        pos = {}
        _child_count = {}
        # For each connected component...
        for _subgraph in S:
            # Calculate a minimum spanning tree and convert it to an undirected graph
            G = nx.to_undirected(nx.minimum_spanning_tree(_subgraph))

            # Process small graphs separately
            if len(G) <= 4:
                as_list = list(G.nodes())
                if len(G) >= 1: pos[as_list[0]] = (0,0)
                if len(G) >= 2: pos[as_list[1]] = (1,1)
                if len(G) >= 3: pos[as_list[2]] = (1,0)
                if len(G) >= 4: pos[as_list[3]] = (0,1)
                continue

            # Determine the root if not set
            my_root = None
            if roots is not None:
                for possible_root in roots:
                    if possible_root in G:
                        my_root = possible_root
            
            # Only set my root to the best if it wasn't already set...
            if my_root is None:
                f = G.copy()
                while len(f) > 2: # while there are more than 2 nodes, remove all the leaves
                    to_be_removed = [x for  x in f.nodes() if f.degree(x) <= 1]
                    f.remove_nodes_from(to_be_removed)
                my_root = list(f)[0]

            # Count the number of children
            _child_count = {}
            for x in G[my_root]: self.__countSubTreeNodes__(G, x, my_root, _child_count)
            root_children_count = 0
            for x in G[my_root]: root_children_count += 1 + _child_count[x]
            _child_count[my_root] = root_children_count

            # Place root
            _leaf_count = {}
            ht_state = HTState(angle=0.0, angle_inc=2.0*pi/self.__totalLeaves__(G, my_root, my_root, _leaf_count), max_depth=self.__treeDepth__(G,my_root,my_root))

            _R_ = 8.0
            def placeChildren(_parent, _node, _depth):
                # Place Leaves Directly
                if _child_count[_node] == 0:
                    pos[_node] = (_depth * _R_ * cos(ht_state.angle) / ht_state.max_depth,
                                  _depth * _R_ * sin(ht_state.angle) / ht_state.max_depth)
                    ht_state.angle += ht_state.angle_inc
                # Interior Node...
                else:
                    begin_angle = ht_state.angle
                    _heap = []
                    for x in G[_node]:
                        if x != _parent: heapq.heappush(_heap, (1/(_child_count[x]+1), x))
                    while len(_heap) > 0:
                        x = heapq.heappop(_heap)[1]
                        placeChildren(_node, x, _depth+1)
                    end_angle = ht_state.angle
                    half_angle = (begin_angle + end_angle)/2
                    pos[_node] = (_depth * _R_ * cos(half_angle) / ht_state.max_depth, _depth * _R_ * sin(half_angle) / ht_state.max_depth)

            for x in G[my_root]: placeChildren(my_root, my_root, 0)
            
            # Touch up center w/ spring layout
            if touch_up_with_springs:
                dists   = dict(nx.all_pairs_shortest_path_length(G))
                leaves  = self.dagLeavesOnly(G)
                centers = set(G.nodes()) - leaves
                pos = self.springLayout(G, pos, centers, iterations=20, spring_exp=0.1, only_sel_adj=True, dists=dists)
                # pos = self.springLayout(G, pos, leaves,  iterations=200, spring_exp=0.1, only_sel_adj=True, dists=dists)

        # Separate the connected components
        if len(S) > 1: return self.treeMapGraphComponentPlacement(_graph,pos,bounds_percent)
        else:          return pos

    #
    # Place children within the hypertree structure
    #
    def __hyperTreePlaceChildren__(self,
                                   pos,
                                   _graph,
                                   _parent,
                                   _node,
                                   _depth,
                                   ht_state,
                                   cen_x,
                                   cen_y,
                                   _child_count,
                                   _leaf_count):
        _R_ = 8.0
        # Place Leaves Directly
        if _child_count[_node] == 0:
            pos[_node] = (cen_x + _depth * _R_ * cos(ht_state.angle) / ht_state.max_depth,
                          cen_y + _depth * _R_ * sin(ht_state.angle) / ht_state.max_depth)
            ht_state.angle += ht_state.angle_inc
        # Interior Node...
        else:
            begin_angle = ht_state.angle
            _heap = []
            for x in _graph[_node]:
                if x != _parent: heapq.heappush(_heap, (1/(_child_count[x]+1), x))
            while len(_heap) > 0:
                x = heapq.heappop(_heap)[1]
                self.__hyperTreePlaceChildren__(pos, _graph, _node, x, _depth+1, ht_state, cen_x, cen_y, _child_count, _leaf_count)
            end_angle = ht_state.angle
            half_angle = (begin_angle + end_angle)/2
            pos[_node] = (cen_x + _depth * _R_ * cos(half_angle) / ht_state.max_depth, cen_y + _depth * _R_ * sin(half_angle) / ht_state.max_depth)


    #
    # treeMapGraphComponentPlacement()
    # - returns a new position map for the graph
    #
    def treeMapGraphComponentPlacement(self, 
                                       _graph,              # graph to place
                                       pos,                 # original positions
                                       bounds_percent=0.1): # border region for the treemap
        # Separate graph into connected components // make sure there are two or more components
        _graph = nx.to_undirected(_graph)
        S = [_graph.subgraph(c).copy() for c in nx.connected_components(_graph)]
        if len(S) <= 1: return pos
        
        # Order the graphs from largest to smallest
        my_order = []
        for _subgraph in S: my_order.append((len(_subgraph),_subgraph))
        my_order.sort(key=lambda x:x[0], reverse=True)

        # Transpose the sizes into an array
        nodes = []
        for tup in my_order: nodes.append(tup[0])

        # Calculate the treemap via the squarify library
        normalized_sizes   = squarify.normalize_sizes(nodes, 1024, 1024)
        treemap_rectangles = squarify.squarify(normalized_sizes,0,0,1024,1024)
        # padded_rectangles  = squarify.padded_squarify(normalized_sizes,0,0,1024,1024) # supposed to add padding...

        # For each subgraph, place within the square
        new_pos = {}
        for i in range(0,len(my_order)):
            _subgraph = my_order[i][1]
            _rect     = treemap_rectangles[i]

            # Adjust for the border percent
            rx,ry,rdx,rdy = _rect['x'],_rect['y'],_rect['dx'],_rect['dy']
            if bounds_percent > 0.0 and bounds_percent < 1.0:
                xperc,yperc = rdx*bounds_percent,rdy*bounds_percent
                rx += xperc/2
                ry += yperc/2
                rdx -= xperc
                rdy -= yperc

            # Determine the current extents
            x0,y0,x1,y1 = self.positionExtents(pos,_subgraph)

            for _node in _subgraph.nodes():
                x = (pos[_node][0] - x0)/(x1-x0)
                y = (pos[_node][1] - y0)/(y1-y0)
                new_pos[_node] = [x*rdx + rx, y*rdy + ry]

        return new_pos

    #
    # randomLayout()
    # - return a random layout.
    #
    def randomLayout(self,_graph):
        pos = {}
        for x in _graph.nodes():
            pos[x] = [random.random(),random.random()]
        return pos

    #
    # jitterLayout()
    # - add jitter to existing layout
    # - mostly used for debugging to separate combined nodes...
    #
    def jitterLayout(self,pos,amount=0.1):
        new_pos = {}
        for k in pos.keys():
            x = pos[k][0]
            y = pos[k][1]
            new_pos[k] = [x + random.random()*amount - amount/2, y + random.random()*amount - amount/2]
        return new_pos
            
    #
    # shortestPathLayout()
    # - for two or more nodes, create a layout that shows the shortest paths for those nodes
    #
    def shortestPathLayout(self,
                           _graph,                  # networkx graph
                           _nodes,                  # list of nodes for the shortest path
                           use_weight     = None,   # parameter for the networkx shortest_path method
                           use_undirected = True):  # use the undirected version of the _graph
        pos    = {}
        placed = set()
        if use_undirected:
            _ugraph = nx.to_undirected(_graph)
        if len(_nodes) >= 2:           
            # Iterative create the shortest path... 
            # ... then  place those nodes... exclude them from the graph... then repeat
            level = 0.0
            _path = []
            while _path is not None:
                # Find the next shortest path and add that positioning
                try:
                    _path = nx.shortest_path(_ugraph, _nodes[0], _nodes[1],weight=use_weight)
                except:
                    _path = None

                if _path is not None:
                    for i in range(0,len(_path)):
                        _node = _path[i]
                        if _node not in placed:
                            x = i/(len(_path)-1)
                            y = (i%2) * 0.1 + level
                            pos[_node] = [x,y]
                            placed.add(_node)
                    
                    # Unfreeze the graph
                    _ugraph = nx.Graph(_ugraph)
                    for _node in placed:
                        if _node in _ugraph.nodes() and _node != _nodes[0] and _node != _nodes[1]:
                            _ugraph.remove_node(_node)

                level += 0.2                    
            
            # Place the remaining... in some type of fashion
            still_not_placed = set()
            for _node in _graph.nodes():
                if _node not in placed:
                    existing = set(x for x in _graph.neighbors(_node)) & placed
                    if len(existing) > 0:
                        x_sum = 0
                        for _nbor in existing:
                            x_sum += pos[_nbor][0]
                        x = x_sum / len(existing)
                        pos[_node] = [x,level]
                        placed.add(_node)
                    else:
                        still_not_placed.add(_node)
            for _node in still_not_placed:
                pos[_node] = [random.random(), level+0.2]

            return pos
        else:
            print(f'shortestPathLayout() requires two or more nodes')
            return self.randomLayout(_graph) # Until implemented
    
    #
    # springLayout()
    # - modeled after the Yet Another Spring Layout java implementation
    # - probably just a reference implementation...  networkx version works much faster...
    #
    def springLayout(self,
                     _graph,               # networkx graph
                     pos          = None,  # If none, will be randomized... otherwise, positions supplied
                                           # will be used as a starting point
                     selection    = None,  # nodes that will be adjusted / None means all nodes
                     only_sel_adj = False, # for each iteration, only use the selection to adjust node positions
                     iterations   = None,  # number of iterations... None means a heuristic will be used
                     dists        = None,  # distance dictionary... if None, will be calculated
                     use_weights  = False, # if true, uses edge weights based on dijkstra's algorithm
                     spring_exp   = 1.0):  # spring exponent
        # Make graph undirected... and separate graph into connected components
        _graph = nx.to_undirected(_graph)
        S = [_graph.subgraph(c).copy() for c in nx.connected_components(_graph)]

        # For each connected component...
        new_pos = {}
        for _subgraph in S:
            # Compute the distance matrix if it hasn't been done already
            if dists is None:
                if use_weights:
                    dists = dict(nx.all_pairs_dijkstra_path_length(_subgraph))
                else:
                    dists = dict(nx.all_pairs_shortest_path_length(_subgraph))

            # How much to move each node
            mu = 1.0/len(_subgraph)

            # Rule of thumb -- spring layout function of the number of nodes
            if iterations is None:
                iterations = len(_subgraph)

            # Determine which nodes the adjustment will apply to
            if selection is None:
                _nodes = set(_subgraph.nodes())
                _space = set(_subgraph.nodes())
            else:
                _nodes = set(_subgraph.nodes()) & set(selection)
                if only_sel_adj:
                    _space = set(_subgraph.nodes()) & set(selection)
                else:
                    _space = set(_subgraph.nodes())
            
            # Initial placement (or copy from the pos parameter)
            for _node in _subgraph.nodes():
                if pos is None or _node not in pos.keys():
                    new_pos[_node] = [random.random(),random.random()]
                else:
                    new_pos[_node] = pos[_node]
                
            # Iterate
            for i in range(0,iterations):
                # Calculate the node adjustment
                x_adj          = {}
                y_adj          = {}
                overall_stress = 0.0
                for _node in _nodes:
                    sum_dx,sum_dy,sum_stress = 0.0,0.0,0.0
                    for _dest in _space:
                        t  = dists[_node][_dest]
                        dx  = new_pos[_node][0] - new_pos[_dest][0]
                        dy  = new_pos[_node][1] - new_pos[_dest][1]
                        dx2 = dx*dx
                        dy2 = dy*dy
                        d   = sqrt(dx2+dy2)
                        e   = pow(t,spring_exp)
                        if d < 0.001:
                            d = 0.001
                        if e < 0.001:
                            e = 0.001
                        sum_dx += (2*dx*(1.0 - t/d))/e
                        sum_dy += (2*dy*(1.0 - t/d))/e
                        sum_stress += (t-d)*(t-d)
                    x_adj[_node] = -mu * sum_dx
                    y_adj[_node] = -mu * sum_dy
                    overall_stress += sum_stress/len(_space)

                # Apply the node adjustment
                for _node in _nodes:
                    new_pos[_node] = [new_pos[_node][0] + x_adj[_node],
                                      new_pos[_node][1] + y_adj[_node]]
        
        return new_pos

    #
    # barycentricLayout()
    # - place the selected nodes 
    #
    def barycentricLayout(self,
                          _graph,                # networkx graph
                          pos,                   # positions of the non-selection
                          selection,             # nodes that will be adjusted
                          dists = None,          # distance dictionary... if None, will be calculated
                          use_weights = False):  # use the weights on the edges
        # Make graph undirected... and separate graph into connected components
        _graph = nx.to_undirected(_graph)
        S = [_graph.subgraph(c).copy() for c in nx.connected_components(_graph)]

        # For each connected component...
        new_pos = {}
        for _subgraph in S:
            # Compute the distance matrix if it hasn't been done already
            if dists is None:
                if use_weights:
                    dists = dict(nx.all_pairs_dijkstra_path_length(_subgraph))
                else:
                    dists = dict(nx.all_pairs_shortest_path_length(_subgraph))
            
            # Pare down to intersection of selection and this subgraph
            to_adjust = selection & _subgraph.nodes()
            fixed     = _subgraph.nodes() - to_adjust
            for _fix in fixed:
                new_pos[_fix] = pos[_fix]

            # Iterate and place
            for _node in to_adjust:
                xsum,ysum,samples = 0,0,0
                for _fix in fixed:
                    xsum += pos[_fix][0] * dists[_node][_fix]
                    ysum += pos[_fix][1] * dists[_node][_fix]
                    samples += 1
                new_pos[_node] = [xsum/samples, ysum/samples]
        
        # Return the new positions
        return new_pos

    #
    # springLayout()
    # - modeled after the Yet Another Spring Layout java implementation
    # - probably just a reference implementation...
    # - threading doesn't really exist unless you want to copy over the complete memory space...
    #
    def springLayoutThreaded(self,
                     _graph,               # networkx graph
                     pos          = None,  # If none, will be randomized... otherwise, positions supplied
                                           # will be used as a starting point
                     selection    = None,  # nodes that will be adjusted / None means all nodes
                     only_sel_adj = False, # for each iteration, only use the selection to adjust node positions
                     iterations   = None,  # number of iterations... None means a heuristic will be used
                     dists        = None,  # distance dictionary... if None, will be calculated
                     use_weights  = False, # if true, uses edge weights based on dijkstra's algorithm
                     spring_exp   = 1.0):  # spring exponent
        #
        # --- vvv --- STRAIGHT COPY FROM REFERENCE --- vvv ---
        #
        # Make graph undirected... and separate graph into connected components
        _graph = nx.to_undirected(_graph)
        S = [_graph.subgraph(c).copy() for c in nx.connected_components(_graph)]

        # For each connected component...
        new_pos = {}
        for _subgraph in S:
            # Compute the distance matrix if it hasn't been done already
            if dists is None:
                if use_weights:
                    dists = dict(nx.all_pairs_dijkstra_path_length(_subgraph))
                else:
                    dists = dict(nx.all_pairs_shortest_path_length(_subgraph))

            # How much to move each node
            mu = 1.0/len(_subgraph)

            # Rule of thumb -- spring layout function of the number of nodes
            if iterations is None:
                iterations = len(_subgraph)

            # Determine which nodes the adjustment will apply to
            if selection is None:
                _nodes = set(_subgraph.nodes())
                _space = set(_subgraph.nodes())
            else:
                _nodes = set(_subgraph.nodes()) & set(selection)
                if only_sel_adj:
                    _space = set(_subgraph.nodes()) & set(selection)
                else:
                    _space = set(_subgraph.nodes())
            
            # Initial placement (or copy from the pos parameter)
            for _node in _subgraph.nodes():
                if pos is None or _node not in pos.keys():
                    new_pos[_node] = [random.random(),random.random()]
                else:
                    new_pos[_node] = pos[_node]

        #
        # --- ^^^ --- STRAIGHT COPY FROM REFERENCE --- ^^^ ---
        #

            # Multi-threaded spring state/holder
            mt_spring = MTSpring(dists, _space, mu, new_pos, spring_exp)

            # Make the nodes into a list for consistency
            node_list = list(_nodes)

            # Iterate
            for i in range(0,iterations):
                # Calculate the node adjustment
                p = Pool(16)
                with p:
                    adjs = p.map(mt_spring.calculateAdjustment, node_list)

                # Apply the node adjustment
                for i in range(0,len(adjs)):
                    _node = node_list[i]
                    new_pos[_node] = [new_pos[_node][0] + adjs[i][0],
                                      new_pos[_node][1] + adjs[i][1]]

        return new_pos


#
# Multi-threaded spring layout... doesn't really work any better...
# Suspect that the issue is that the cost of creating the processes
# is too high -- i.e., each has to get a copy of the distance matrix
# and the graph...
#
class MTSpring(object):
    #
    # Constructor
    #
    def __init__(self, 
                 dists,        # distance calculation 
                 space,        # space to adjust over
                 mu,           # how much to adjust each node
                 pos,          # current positions of the nodes
                 spring_exp):  # spring exponent 
        self.dists      = dists
        self.space      = space
        self.mu         = mu
        self.pos        = pos
        self.spring_exp = spring_exp
    
    #
    # Calculate the adjustment for a single node
    #
    def calculateAdjustment(self, node):
        sum_dx,sum_dy,sum_stress = 0.0,0.0,0.0
        for _dest in self.space:
            t  = self.dists[node][_dest]
            dx  = self.pos[node][0] - self.pos[_dest][0]
            dy  = self.pos[node][1] - self.pos[_dest][1]
            dx2 = dx*dx
            dy2 = dy*dy
            d   = sqrt(dx2+dy2)
            e   = pow(t,self.spring_exp)
            if d < 0.001:
                d = 0.001
            if e < 0.001:
                e = 0.001
            sum_dx += (2*dx*(1.0 - t/d))/e
            sum_dy += (2*dy*(1.0 - t/d))/e
            sum_stress += (t-d)*(t-d)
        #    
        #      x_adj,     y_adj,     stress
        #
        return -self.mu*sum_dx,-self.mu*sum_dy,sum_stress/len(self.space) 

#
# HyperTree state holder/struct
#
@dataclass
class HTState:
    angle:     float
    angle_inc: float
    max_depth: int





# Copyright 2023 David Trimm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pandas as pd
import numpy as np

__name__ = 'rt_datamanip_mixin'

#
# Data Manipulation Mixin
# ... utilities for preparing a dataframe for the visualization components
#
class RTDataManipMixin(object):
    #
    # temporalStatsAggregation()
    # ... Produces a variety of stats based on specified temporal frequency, a list of fields, and stat names
    # ... Returns as a pandas dataframe with field_stat columns ... index is the temporal aggregation
    #
    # ... and yes, there's probably something that already does this in the pandas library...
    #
    def temporalStatsAggregation(self, df, ts_field=None, freq='YS', fields=[], stats=['sum','max','median','mean','min','stdev','rows','set_size']):
        # Convert parameters to a list if necessary
        if type(fields) != list:
            fields = [fields]
        if type(stats)  != list:
            stats  = [stats]

        # Determine the timestamp field
        if ts_field is None:
            ts_field = self.guessTimestampField(df)

        # Determine if a field is a categorical type (for set-based operations only)
        field_is_set = {}
        for field in fields:
            field_is_set[field] = self.countBySet(df, field)

        # Operations that can only be done if the field is all numbers
        numeric_ops = ['sum','max','median','mean','min','stdev']

        # Initialize the column contain
        _lu = {}
        for field in fields:
            for stat in stats:
                if stat in numeric_ops and field_is_set[field] == False:
                    _lu[field + '_' + stat] = []
                elif stat not in numeric_ops:
                    _lu[field + '_' + stat] = []

        # Produce the stats
        indices = []
        gb = df.groupby(pd.Grouper(key=ts_field, freq=freq))
        for k,k_df in gb:
            indices.append(k)
            for field in fields:
                # ================================================================= #
                if 'sum' in stats and field_is_set[field] == False:
                    _lu[field + '_sum']     .append(k_df[field].sum())
                if 'max' in stats and field_is_set[field] == False:
                    _lu[field + '_max']     .append(k_df[field].max())
                if 'median' in stats and field_is_set[field] == False:
                    _lu[field + '_median']  .append(k_df[field].median())
                if 'mean' in stats and field_is_set[field] == False:
                    _lu[field + '_mean']    .append(k_df[field].mean())
                if 'min' in stats and field_is_set[field] == False:
                    _lu[field + '_min']     .append(k_df[field].min())
                if 'stdev' in stats and field_is_set[field] == False:
                    _lu[field + '_stdev']   .append(k_df[field].std())

                # ================================================================= #
                if 'rows' in stats:
                    _lu[field + '_rows']    .append(len(k_df))
                if 'set_size' in stats:
                    _lu[field + '_set_size'].append(len(set(k_df[field])))

        # Return the dataframe
        _df = pd.DataFrame(_lu, index=indices)
        _df.index.name = ts_field
        return _df

    #
    # temporalStatsAggregationWithGBFields()
    # ... same as above but keeps the gb_fields separable
    #
    def temporalStatsAggregationWithGBFields(self, 
                                             df,                   # Dataframe to aggregate
                                             fields,               # Field or fields to aggregate
                                             ts_field=None,        # timestamp field... if none, method will use first one found
                                             gb_fields=[],         # Fields to keep separable
                                             flatten_index=True,   # Flatten the index before returning the aggregation
                                             fill_missing=False,   # Fill in missing timestamps 
                                             freq='YS',            # Frequency for the aggregation
                                             stats=['sum','max','median','mean','min','stdev','rows','set_size']):
        # Convert parameters to a list if necessary
        if type(fields) != list:
            fields = [fields]
        if type(stats)  != list:
            stats  = [stats]
        if type(gb_fields) != list:
            gb_fields = [gb_fields]

        # Determine the timestamp field
        if ts_field is None:
            ts_field = self.guessTimestampField(df)

        # Determine if a field is a categorical type (for set-based operations only)
        field_is_set = {}
        for field in fields:
            field_is_set[field] = self.countBySet(df, field)

        # Operations that can only be done if the field is all numbers
        numeric_ops = ['sum','max','median','mean','min','stdev']

        # Initialize the column contain
        _lu = {}
        for field in fields:
            for stat in stats:
                if stat in numeric_ops and field_is_set[field] == False:
                    _lu[field + '_' + stat] = []
                elif stat not in numeric_ops:
                    _lu[field + '_' + stat] = []

        # Produce the stats
        indices     = []
        complete_gb,complete_index = [],[]
        complete_gb.    append(pd.Grouper(key=ts_field, freq=freq))
        complete_index. append(ts_field)
        for x in gb_fields:
            complete_gb.   append(x)
            complete_index.append(x)

        earliest_seen,latest_seen = None,None

        tuples_seen = set()
        gb = df.groupby(complete_gb)
        for k,k_df in gb:

            # Keep earliest & latest
            if earliest_seen is None:
                earliest_seen = k[0]
            latest_seen=k[0]

            # Keep track of tuples seen and the separate index
            tuples_seen.add(k)
            indices.append(k)

            # Calculate the stats
            for field in fields:
                # ================================================================= #
                if 'sum' in stats and field_is_set[field] == False:
                    _lu[field + '_sum']     .append(k_df[field].sum())
                if 'max' in stats and field_is_set[field] == False:
                    _lu[field + '_max']     .append(k_df[field].max())
                if 'median' in stats and field_is_set[field] == False:
                    _lu[field + '_median']  .append(k_df[field].median())
                if 'mean' in stats and field_is_set[field] == False:
                    _lu[field + '_mean']    .append(k_df[field].mean())
                if 'min' in stats and field_is_set[field] == False:
                    _lu[field + '_min']     .append(k_df[field].min())
                if 'stdev' in stats and field_is_set[field] == False:
                    _lu[field + '_stdev']   .append(k_df[field].std())

                # ================================================================= #
                if 'rows' in stats:
                    _lu[field + '_rows']    .append(len(k_df))
                if 'set_size' in stats:
                    _lu[field + '_set_size'].append(len(set(k_df[field])))

        # Fill in missing values
        if fill_missing:
            if len(gb_fields) == 1:
                gb_separable = df.groupby(gb_fields[0])
            else:
                gb_separable = df.groupby(gb_fields)
            for k,k_df in gb_separable:
                for _date in pd.date_range(start=earliest_seen, end=latest_seen, freq=freq):
                    _k_as_list = list()
                    _k_as_list.append(_date)
                    if type(k) == tuple:
                        _k_as_list += k
                    else:
                        _k_as_list.append(k)
                    _tuple = tuple(_k_as_list)
                    if _tuple not in tuples_seen:
                        indices.append(_tuple)
                        for x in _lu.keys():
                            _lu[x].append(0)

        # Create the dataframe
        _df = pd.DataFrame(_lu, index=indices)

        # Flatten the index if requested
        if flatten_index:
            _df = _df.reset_index()
            _df = _df.join(pd.DataFrame(_df['index'].values.tolist(), columns=complete_index))\
                     .drop('index',axis=1)

        return _df

    #
    # rowContainsSubstring()
    # - use it as follows:  df[df.apply(lambda x: rt.rowContainsSubstring(x, 'sub'),axis=1)]
    #
    def rowContainsSubstring(self,
                             _row,
                             _substring,
                             _match_case=False):
        if _match_case == False:
            _substring = _substring.lower()
        for x in _row:
            if _match_case == False:
                x = str(x).lower()
            if _substring in x:
                return True
        return False
    


# Copyright 2023 David Trimm
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# tokenizers             0.5.2
# torch                  2.0.1+cu118
# Transformers           2.8.0

import pandas as pd
import numpy as np
from numpy.linalg import norm
import re

import networkx as nx # for TextRank

from rt_component import RTComponent # Unused?

from shapely.geometry import Polygon, MultiPolygon
import shapely.affinity as affinity

__name__ = 'rt_text_mixin'

#
# Abstraction for Text
#
class RTTextMixin(object):
    #
    # Constructor
    # 
    def __text_mixin_init__(self):
        self.spacy_loaded_flag = False

    #
    # textBlock() - render a textblock and track positional information of characters and words.
    #
    def textBlock(self,
                  txt,
                  txt_h=14,
                  line_space_px=3,
                  word_wrap=False,
                  w=512,
                  x_ins=5,
                  y_ins=3):
        
        svg,x,y,line = '',x_ins,y_ins+txt_h,''
        last_was_space = True
        line_lu = {}

        for i in range(0,len(txt)):
            c = txt[i]
            if c == '\n':
                # CODE BLOCK B
                line_lu[y]     = line + '\n'
                x              = x_ins
                y             += txt_h + line_space_px
                line           = ''
                last_was_space = True
                if word_wrap:
                    y += 3*line_space_px
            elif word_wrap == False:
                # CODE BLOCK A
                line          += c
                x             += self.textLength(c,txt_h)
            else:
                if last_was_space and self.__whitespace__(c) == False:
                    j,x_j = i+1,x+self.textLength(c,txt_h)
                    while j < len(txt) and self.__whitespace__(txt[j]) == False:
                        x_j += self.textLength(txt[j],txt_h)
                        j   += 1
                    if   x_j > (w-x_ins) and x != x_ins: # new word exceeds the maximum width / start new line
                        # CODE BLOCK B
                        line_lu[y]     = line
                        x              = x_ins
                        y             += txt_h + line_space_px
                        line           = ''
                        last_was_space = True
                        # CODE BLOCK A
                        line          += c
                        x             += self.textLength(c,txt_h)
                    elif (x_j > (w-x_ins) and x == x_ins) or ((x_j - x) >= (w-x_ins)): # a chunk of text is too long to fit on a line
                        # CODE BLOCK B-mod
                        line_lu[y]     = line
                        x              = x_ins
                        line           = ''
                        last_was_space = False # MOD HERE
                        # CODE BLOCK A
                        line          += c
                        x             += self.textLength(c,txt_h)                        
                    else:                        # fine to add the word
                        # CODE BLOCK A
                        line          += c
                        x             += self.textLength(c,txt_h)
                else:
                        # CODE BLOCK A
                        line          += c
                        x             += self.textLength(c,txt_h)
                if self.__whitespace__(c):
                    last_was_space = True

        # If there's a left over line, add it here...
        if len(line) > 0:
            line_lu[y]   = line
            y           += txt_h + line_space_px

        # Just render as is...
        _ignore_ = """
        orig_to_xy,i = {},0
        for k in line_lu:
            line = line_lu[k]
            svg += self.svgText(line_lu[k], x_ins, k, txt_h, color='#b0b0b0')
            x = x_ins
            for c in line:
                orig_to_xy[i] = (x,k)
                x += self.textLength(c, txt_h)
                i += 1
        """

        # Render each word individually -- w/ absolute coordinates
        orig_to_xy,i = {},0
        for y in line_lu:
            line = line_lu[y]
            x,j  = x_ins,0
            while j < len(line):
                c = line[j]
                svg += self.svgText(c, x, y, txt_h)
                orig_to_xy[i] = (x,y)
                j += 1
                i += 1
                x += self.textLength(c, txt_h)
        y += txt_h + line_space_px

        # Calculate geom_to_word
        geom_to_word = {}
        i,last_was_space = 0,True
        _dn = 4 # downward shift...
        while i < len(txt):
            if self.__whitespace__(txt[i]) or self.__punctuation__(txt[i]):
                last_was_space = True
                i += 1
            else:
                if last_was_space and self.__whitespace__(txt[i]) == False and self.__punctuation__(txt[i]) == False:
                    i0 = i
                    while i < len(txt)                           and \
                          self.__whitespace__ (txt[i])  == False and \
                          self.__punctuation__(txt[i])  == False:
                        i += 1
                        i1 = i
                    x0,y0 =  orig_to_xy[i0]
                    x1,y1 =  orig_to_xy[i1-1]
                    x1    += self.textLength(txt[i-1],txt_h)
                    _polygon = Polygon([[x0,y0+line_space_px+_dn], 
                                        [x1,y1+line_space_px+_dn], 
                                        [x1,y1-txt_h+_dn], 
                                        [x0,y1-txt_h+_dn]])
                    geom_to_word[_polygon] = txt[i0:i1]
                last_was_space = False
                i = i1

        bounds = (0,0,w,y-txt_h+y_ins)
        return RTTextBlock(self, txt, txt_h, line_space_px, word_wrap, w, x_ins, y_ins, svg, bounds, geom_to_word, orig_to_xy)

    # Is character whitespace?
    def __whitespace__ (self, c):
        return c == ' ' or c == '\t' or c == '\n'
    
    # Is character punctuation?
    def __punctuation__(self, c):
        _str = '''!.?,[]{}:;`~%^&*()-_+='"<>/\\'''
        return c in _str
    
    #
    # textJoinLines() - join lines together and remove extra spaces.
    # - expect that this is a utility to call before textBlock()
    #
    def textJoinNewLines(self, txt):
        joined = ' '.join(txt.split('\n'))
        while len(joined) > 0 and joined[0] == ' ':
            joined = joined[1:]
        while len(joined) > 0 and joined[-1] == ' ':
            joined = joined[:-1]
        words = joined.split(' ')
        wout_blanks = []
        for word in words:
            if len(word) > 0:
                wout_blanks.append(word)
        return ' '.join(wout_blanks)
    
    #
    # textJoinNewLinesBetter() - keep newlines (if single) intact...
    # - more closely mirrors the (de facto) standard of using double line returns 
    #   to separate paragraphs
    # - expect that this is a utility to call before textBlock()
    #
    def textJoinNewLinesBetter(self, txt):
        re_match = re.findall(r'([\n]{2,})',txt)
        if re_match is None:
            return self.joinNewLines(txt)
        else:
            i,_ret = 0,''
            for _match in re_match:
                j = txt.find(_match,i)
                _ret += self.textJoinNewLines(txt[i:j])
                for k in range(len(_match)-1):
                    _ret += '\n'
                i = j+len(_match)
            _ret += self.textJoinNewLines(txt[i:])
            return _ret
    
    #
    # maxLinePixels() - split a string by new line characters, then determine
    # the maximum line length (in pixels).
    #
    def textMaxLinePixels(self, txt, txt_h=14):
        _max = 0
        lines = txt.split('\n')
        for _line in lines:
            _len = self.textLength(_line, txt_h)
            _max = max(_len,_max)
        return _max + 6

    #
    # __loadSpacy__():  Load spacy (only if necessary)
    #
    def __loadSpacy__(self):
        if self.spacy_loaded_flag == False:
            import spacy
            self.nlp_spacy = spacy.load('en_core_web_sm')
            self.spacy_loaded_flag = True

    #
    # textSubsequenceLookup() - find all subsequences with length at least min_length.
    # ... sequences are delimited by spaces and punctuations
    # ... punctuations are included in the sequence... spaces are not
    # ... based off of dynamic programming pseudocode in the following:
    # ... ... https://en.wikipedia.org/wiki/Longest_common_substring
    # ... this appears to miss the same occurence of the sequence if it occurs in different places...
    #
    def textSubsequenceLookup(self,
                              s,
                              t,
                              min_length=4):       
        s_toks, t_toks = self.__textSubsequenceLookup_as_tokens__(s), self.__textSubsequenceLookup_as_tokens__(t)
        # s_toks_str,t_toks_str = list(zip(*s_toks))[0],list(zip(*t_toks))[0]
        r,n = len(s_toks),len(t_toks)
        z   = 0
        ret = ''
        L   = {}
        for i in range(r):
            L[i] = {}
        for i in range(r):
            for j in range(n):
                if s_toks[i][0] == t_toks[j][0]:
                    if i == 0 or j == 0:
                        L[i][j] = 0
                    else:
                        L[i][j] = L[i-1][j-1] + 1
                    if L[i][j] > z:
                        z   = L[i][j]
                else:
                    L[i][j] = 0
        results = {}
        for i in L.keys():
            for j in L[i].keys():
                if L[i][j] >= min_length:
                    no_longer_builds = True
                    if i < (r-1) and j < (n-1) and L[i+1][j+1] == L[i][j]+1:
                        no_longer_builds = False
                    if no_longer_builds:
                        l = L[i][j]
                        i0 = i - l + 1
                        j0 = j - l + 1
                        if l not in results.keys():
                            results[l] = []
                        results[l].append(((s_toks[i0][1], s_toks[i0+l-1][2]),(t_toks[j0][1], t_toks[j0+l-1][2])))
        return results

    #
    # Supporting method for the subsequence lookup...
    #
    def __textSubsequenceLookup_as_tokens__(self, s, keep_punctuation=False):
        i,i0,results = 0,None,[]
        while i < len(s):
            if   self.__whitespace__(s[i]):
                if i0 is not None:
                    results.append((s[i0:i],i0,i))
                    i0 = None
            elif self.__punctuation__(s[i]):
                if i0 is not None:                    
                    results.append((s[i0:i],i0,i))
                    i0 = None
                if keep_punctuation:
                    results.append((s[i:i+1],i,i+1))
            else:
                if i0 is None:
                    i0 = i
            i += 1
        if i0 is not None:
            results.append((s[i0:i],i0,i))
            i0 = None
        return results

    #
    # textCreateHighlightsLookupBasedOnSubsequenceResults() - create a higlights dictionary for the results
    #
    def textCreateHighlightsLookupBasedOnSubsequenceResults(self,
                                                            _text,     # text passed into the textSequenceLookup() call
                                                            _results,  # return of the textSequenceLookup() call
                                                            _index):   # should either by 0 or 1
        # Because the subsequence calc doesn't use punctuation... the actual string needs to be minimized to remove spacing and punctuations
        def noSpacesOrPunctuations(s):
            r = ''
            for i in range(len(s)):
                if self.__whitespace__(s[i]) == False and self.__punctuation__(s[i]) == False:
                    r += s[i]
            return r

        _lu = {}
        for k in _results.keys():
            _list = _results[k]
            for _tuple in _list:
                i0,i1 = _tuple[_index][0],_tuple[_index][1]
                _lu[_tuple[_index]] = self.co_mgr.getColor(noSpacesOrPunctuations(_text[i0:i1]))
        return _lu

    #
    # textExtractSentences() - extract sentences using spacy
    #
    # _tups = textExtractSentences(_str_)
    # _just_the_sentences_as_array = list(list(zip(*_tups))[0])
    #
    def textExtractSentences(self,
                             txt,
                             split_by_newlines=True):
        parts = txt.split('\n') if split_by_newlines else [txt]
        self.__loadSpacy__()
        sentences = []
        k = 0
        for part in parts:
            i = 0
            for _span in self.nlp_spacy(part).sents:
                as_str = str(_span)
                while len(as_str) >0 and self.__whitespace__(as_str[0]):
                    as_str = as_str[1:]
                if len(as_str) > 0:
                    i      = part.index(as_str,i)
                    j      = i + len(as_str)
                    while j < len(part) and part[j] == ' ':
                        as_str += ' '
                        j      += 1
                    sentences.append((as_str, i+k, j+k))
                    i += len(as_str)
            k += len(part) + 1            
        return sentences

    #
    # textExtractSentences() - extract sentences
    # - original version in NLTK... but rewrote in spacy...
    #
    # _tups = textExtractSentences(_str_)
    # _just_the_sentences_as_array = list(list(zip(*_tups))[0])
    #
    def __textExtractSentences_NLTK_version__(self,
                                          txt):
        import nltk
        tokens,sentences = nltk.sent_tokenize(txt),[]
        if len(tokens) > 0:
            i = txt.index(tokens[0])
        for _token in tokens:
            i = txt.index(_token,i)
            sentences.append((_token, i, i + len(_token)))
            i += len(_token)
        return sentences

    #
    # textExtractEntities() - extract entities.
    #
    def textExtractEntities(self, 
                            txt, 
                            algo='spacy'):
        if algo == 'spacy':
            return self.__textExtractEntitiesSpacy__(txt)
        else:
            raise Exception(f'RACETrack.textExtractEntities() - unknown algorithm "{algo}"')

    #
    # __extractEntitiesSpacy__() - extract entities using SpaCy.
    #
    def __textExtractEntitiesSpacy__(self,txt):
        self.__loadSpacy__()        
        doc = self.nlp_spacy(txt)
        ret = []
        for entity in doc.ents:
            ret.append((entity.text, entity.label_, entity.end_char - len(entity.text), entity.end_char))
        return ret

    #
    # textLexRank()
    # - implemented poorly :(
    #
    def textLexRank(self, txt, embed_fn):
        self.__loadSpacy__()
        sentence_tuples = self.textExtractSentences(txt)
        _zipped         = list(zip(*sentence_tuples))
        sentences       = _zipped[0]
        i0s             = _zipped[1]
        i1s             = _zipped[2]
        sentence_vecs   = embed_fn(sentences)
        g_nx            = nx.Graph()
        for i in range(len(sentence_vecs)):
            for j in range(len(sentence_vecs)):
                if i != j:
                    _sim = np.dot(sentence_vecs[i], sentence_vecs[j])/(norm(sentence_vecs[i])*norm(sentence_vecs[j]))
                    g_nx.add_edge(i,j,weight=_sim)
        pagerank = nx.pagerank(g_nx)
        scores    = []
        for i in range(len(sentence_vecs)):
            scores.append(pagerank[i])
        df = pd.DataFrame({'sentence':sentences,'i0':i0s,'i1':i1s,'lr_score':scores})
        return df

    #
    # textRank()
    # - modeled after https://github.com/davidadamojr/TextRank
    # -- which appears modeled after https://web.eecs.umich.edu/~mihalcea/papers/mihalcea.emnlp04.pdf
    #
    def textRank(self, txt):
        self.__loadSpacy__()
        sentence_tuples = self.textExtractSentences(txt)
        sentences,begs,ends,non_stops = [],[],[],[]
        for _tuple in sentence_tuples:
            sentence = _tuple[0]
            sentence_nlp = self.nlp_spacy(sentence)
            for token in sentence_nlp:
                if token.is_stop == False and self.__punctuation__(str(token)) == False:
                    sentences.append(sentence)
                    begs.     append(_tuple[1])
                    ends.     append(_tuple[2])
                    non_stops.append(str(token))
        df             = pd.DataFrame({'sentence':sentences, 'i0':begs, 'i1':ends, 'non_stops':non_stops})
        relationships  = [("sentence","non_stops")]
        g_nx           = self.createNetworkXGraph(df, relationships)
        pagerank       = nx.pagerank(g_nx)
        df['pr_score'] = df['sentence'].apply(lambda x: pagerank[x])
        return df

    #
    # textSummaryHeatmap()
    #
    def textSummaryHeatmap(self,
                           text_main,
                           text_summary,
                           embed_fn,
                           collapse=False,       # collapse to a single row
                           cell_w=None,          # cell width <== use to override the w of the overall SVG markup
                           cell_h=None,          # cell height <== use to overrid the h of the overall SVG markup
                           global_dot_min=None,
                           global_dot_max=None,
                           x_gap = 1,
                           y_gap = 1,
                           x_ins = 3,
                           y_ins = 3,
                           w=128,
                           h=32):
        _main_sentences               = self.textExtractSentences(text_main)
        _main_sentences_embeddings    = embed_fn(list(zip(*_main_sentences))[0])
        _summary_sentences            = self.textExtractSentences(text_summary)
        _summary_sentences_embeddings = embed_fn(list(zip(*_summary_sentences))[0])
        if cell_w is None:
            cell_w = (w-2*x_ins) / len(_main_sentences)
            x_gap  = 0
        else:
            w = 2*x_ins + cell_w * len(_main_sentences) + x_gap * (len(_main_sentences) - 1)

        if cell_h is None:
            cell_h = (h-2*y_ins) / len(_summary_sentences)
            y_gap  = 0
        else:
            h = 2*y_ins + cell_h * len(_summary_sentences) + y_gap * (len(_summary_sentences) - 1)

        rows,_dot_min,_dot_max = [],None,None
        for i in range(len(_summary_sentences)):
            row = []
            for j in range(len(_main_sentences)):
                _dot = float(np.tensordot(_summary_sentences_embeddings[i], _main_sentences_embeddings[j], axes=1)) # Works with Google's Universal Sentence Embedder...
                if _dot_min is None:
                    _dot_min,_dot_max = _dot,_dot
                _dot_min = min(_dot, _dot_min)
                _dot_max = max(_dot, _dot_max)
                row.append(_dot)
            rows.append(row)

        if global_dot_min is not None:
            _dot_min = global_dot_min
        if global_dot_max is not None:
            _dot_max = global_dot_max

        if collapse:
            h = 2*y_ins + cell_h
            svg = f'<svg x="0" y="0" width="{w}" height="{h}">'
            _co = '#000000' # self.co_mgr.getTVColor("background","default")
            svg += f'<rect x="{0}" y="{0}" width="{w}" height="{h}" fill="{_co}" />'
            y   = y_ins    
            for xi in range(len(_main_sentences)):
                x = x_ins + xi * (cell_w + x_gap)
                _cell_min = _cell_max = rows[0][xi]
                for yi in range(len(_summary_sentences)):
                    _cell_min = min(_cell_min, rows[yi][xi])
                    _cell_max = max(_cell_max, rows[yi][xi])
                if _cell_max > _dot_max:
                    _co  = '#ff0000'
                else:
                    _hex =  int(min(255,int(255.0*(_cell_max - _dot_min)/(_dot_max - _dot_min))))
                    _co  =  f'#{_hex:02x}{_hex:02x}{_hex:02x}'
                svg  += f'<rect x="{x}" y="{y}" width="{cell_w}" height="{cell_h}" fill="{_co}" />'
                if _cell_min < _dot_min:
                    _co = '#0000ff'
                else:
                    _hex =  int(min(255,int(255.0*(_cell_min - _dot_min)/(_dot_max - _dot_min))))
                    _co  =  f'#{_hex:02x}{_hex:02x}{_hex:02x}'
                svg  += f'<path d="M {x} {y+cell_h} L {x+cell_w} {y+cell_h} L {x+cell_w} {y} Z" fill="{_co}" />'
            svg += '</svg>'
            return svg
        else:
            svg = f'<svg x="0" y="0" width="{w}" height="{h}">'
            _co = '#000000' # self.co_mgr.getTVColor("background","default")
            svg += f'<rect x="{0}" y="{0}" width="{w}" height="{h}" fill="{_co}" />'
            for yi in range(len(_summary_sentences)):
                y = y_ins + yi * (cell_h + y_gap)
                for xi in range(len(_main_sentences)):
                    x = x_ins + xi * (cell_w + x_gap)
                    if   rows[yi][xi] > _dot_max:
                        _co = '#ff0000'
                    elif rows[yi][xi] < _dot_min:
                        _co = '#0000ff'
                    else:
                        _hex =  int(min(255,int(255.0*(rows[yi][xi] - _dot_min)/(_dot_max - _dot_min))))
                        _co  =  f'#{_hex:02x}{_hex:02x}{_hex:02x}'
                    svg  += f'<rect x="{x}" y="{y}" width="{cell_w}" height="{cell_h}" fill="{_co}" />'
            svg += '</svg>'
            return svg

    #
    # textCompareSummaries()
    #
    # Methodologies include
    #   "sentence_embeddings"
    #   "sentence_embeddings_pixels"
    #   "bert_top_n_prob"
    #   "missing_words"
    #   "longest_common_subsequence"
    #
    def textCompareSummaries(self, 
                             text_main,
                             text_summaries,
                             methodology      = "sentence_embeddings", 

                             # SENTENCE EMBEDDINGS
                             embed_fn         = None,                    # For the two "sentence_embeddings" methods

                             # BERT_TOP_N
                             model            = None,                    # For the two "bert_top_n" methods
                             tokenizer        = None,                    # For the two "bert_top_n" methods
                             device           = None,                    # For the two "bert_top_n" methods

                             # LONGEST_COMMON_SUBSEQUENCE
                             min_length       = 4,                       # For the "longest_common_subsequence" method

                             # Standard Rendering Params
                             main_txt_h       = 14,
                             summary_txt_h    = 16,
                             spacing          = 16,
                             opacity          = 0.8,
                             w                = 1280):
        if type(text_summaries) == str:
            text_summaries = {'Default':text_summaries}
        if   methodology == "sentence_embeddings":
            return self.__textCompareSummaries__sentence_embeddings__(text_main, text_summaries, embed_fn, main_txt_h, summary_txt_h, spacing, opacity, w)
        elif methodology == "sentence_embeddings_pixels":
            return self.__textCompareSummaries__sentence_embeddings_pixels__(text_main, text_summaries, embed_fn, summary_txt_h, spacing, w)
        elif methodology == "bert_top_n" or methodology == "bert_top_n_prob":
            return self.__textCompareSummaries__bert_top_n__(text_main, text_summaries, methodology, model, tokenizer, device, main_txt_h, summary_txt_h, spacing, opacity, w)
        elif methodology == "missing_words":
            return self.__textCompareSummaries__missing_words__(text_main, text_summaries, main_txt_h, summary_txt_h, spacing, opacity, w)
        elif methodology == "longest_common_subsequence":
            return self.__textCompareSummaries__longest_common_subsequence__(text_main, text_summaries, min_length, main_txt_h, summary_txt_h, spacing, opacity, w)
        else:
            raise Exception(f'RACETrack.textCompareSummaries() - unknown methodology "{methodology}"')

    #
    # __textCompareSummaries__longest_common_subsequence__()
    #
    def __textCompareSummaries__longest_common_subsequence__(self,
                                                             text_main, 
                                                             text_summaries, 
                                                             min_length,
                                                             main_txt_h,
                                                             summary_txt_h, 
                                                             spacing,
                                                             opacity,
                                                             w):
        summary_w = main_w = (w - spacing)/2
        _svgs         = []
        _summary_svgs = []
        main_rttb = self.textBlock(text_main, txt_h=main_txt_h, line_space_px=3+3*len(text_summaries.keys()), word_wrap=True, w=main_w)
        main_underlines_svg = ''
        summary_i = 0
        for _summary_desc in text_summaries:
            _summary             = text_summaries[_summary_desc]
            _rttb                = self.textBlock(_summary, txt_h=summary_txt_h, word_wrap=True, w=summary_w)
            _results             = self.textSubsequenceLookup(text_main, _summary, min_length=min_length)
            _summary_lu          = self.textCreateHighlightsLookupBasedOnSubsequenceResults(_summary,  _results, 1)
            _main_lu             = self.textCreateHighlightsLookupBasedOnSubsequenceResults(text_main, _results, 0)
            main_underlines_svg += main_rttb.underlinesOverlay(_main_lu, y_offset=3*summary_i, underline_stroke_w=4)
            _summary_svgs.append(f'<svg x="0" y="0" width="{summary_w}" height="{24}">' + \
                                 f'<rect x="0" y="0" width="{summary_w}" height="{24}" fill="#000000" />' + \
                                 self.svgText(_summary_desc, 3, 20, txt_h=19, color='#ffffff') + '</svg>')
            _summary_svgs.append(_rttb.underlines(_summary_lu, underline_stroke_w=4))
            _summary_svgs.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"></svg>')
            summary_i += 1

        return self.tile([self.tile(_summary_svgs, horz=False)._repr_svg_(),
                          f'<svg x="0" y="0" width="{spacing}" height="{spacing}"></svg>',
                          main_rttb.wrap(main_rttb.background() + main_rttb.unwrappedText() + main_underlines_svg)])

    #
    # __textCompareSummaries__sentence_embeddings_pixels__()
    #
    def __textCompareSummaries__sentence_embeddings_pixels__(self,
                                                             text_main, 
                                                             text_summaries, 
                                                             embed_fn, 
                                                             summary_txt_h, 
                                                             spacing,
                                                             w):
        _svgs                      = []
        _main_sentences            = self.textExtractSentences(text_main)
        _main_sentences_embeddings = embed_fn(list(zip(*_main_sentences))[0])
        _main_rttb                 = self.textBlock(text_main,txt_h=summary_txt_h,w=w,word_wrap=True)

        _svgs.append(f'<svg x="0" y="0" width="{w}" height="{24}">' + \
                     f'<rect x="0" y="0" width="{w}" height="{24}" fill="#000000" />' + \
                     self.svgText('Main', 3, 20, txt_h=19, color='#ffffff') + '</svg>')

        _svgs.append(_main_rttb._repr_svg_())
        _svgs.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"></svg>')
        for _desc in text_summaries.keys():
            _summary                      = text_summaries[_desc]
            _summary_sentences            = self.textExtractSentences(_summary)
            _summary_sentences_embeddings = embed_fn(list(zip(*_summary_sentences))[0])
            # Calculate all comparisons... 
            _dot_min,_dot_max,_dots       = None,None,{}
            for i in range(len(_main_sentences)):            
                _dots[i] = {}
                for j in range(len(_summary_sentences)):
                    _dot = float(np.tensordot(_summary_sentences_embeddings[j], _main_sentences_embeddings[i], axes=1)) # Works with Google's Universal Sentence Embedder...
                    if _dot_min is None:
                        _dot_min = _dot
                    _dot_min = min(_dot, _dot_min)
                    if _dot_max is None:
                        _dot_max = _dot
                    _dot_max = max(_dot, _dot_max)
                    _dots[i][j] = _dot
            # Rendering ... per summary sentences
            _summary_svgs = []
            _summary_svgs.append(self.textBlock(text_summaries[_desc], txt_h=summary_txt_h, word_wrap=True, w=w)._repr_svg_())
            _summary_svgs.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"></svg>')
            _summary_svgs.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"></svg>')
            for j in range(len(_summary_sentences)):
                _highlights = {}
                for i in range(len(_main_sentences)):
                    _hex = 255 - int(min(255,int(255.0*(_dots[i][j] - _dot_min)/(_dot_max - _dot_min))))
                    _highlights[(_main_sentences[i][1],_main_sentences[i][2])] = f'#{_hex:02x}{_hex:02x}{_hex:02x}'
                _summary_svgs.append(_main_rttb.pixelRepr(_highlights, w/3))
                _summary_svgs.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"></svg>')
            _svgs.append(f'<svg x="0" y="0" width="{w}" height="{24}">' + \
                         f'<rect x="0" y="0" width="{w}" height="{24}" fill="#000000" />' + \
                         self.svgText(_desc, 3, 20, txt_h=19, color='#ffffff') + '</svg>')
            _svgs.append(self.tile(_summary_svgs, horz=True)._repr_svg_())
            _svgs.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"></svg>')
        return self.tile(_svgs, horz=False)

    #
    # __textCompareSummaries__missing_words__()
    #
    def __textCompareSummaries__missing_words__(self,
                                                text_main,
                                                text_summaries,
                                                main_txt_h,
                                                summary_txt_h,
                                                spacing,
                                                opacity,
                                                w):
        # Geometry
        main_w = summary_w = (w - spacing)/2

        # Extract words from a text
        def words(_txt):
            i,i0,_set = 0,None,set()
            while i < len(_txt):
                if self.__whitespace__(_txt[i]) or self.__punctuation__(_txt[i]):
                    if i0 is not None:
                        _set.add(_txt[i0:i].lower())
                        i0 = None
                else:
                    if i0 is None:
                        i0 = i
                i += 1
            if i0 is not None:
                _set.add(_txt[i0:i].lower())
            return _set

        # Create sets of the words found in each
        main_words    = words(text_main)
        summary_words = set()
        for summary_desc in text_summaries:
            summary_words |= words(text_summaries[summary_desc])

        # Highlight words not found in the _set
        def highlightsForText(_txt, _set, _co):
            highlights = {}
            i,i0 = 0,None
            while i < len(_txt):
                if self.__whitespace__(_txt[i]) or self.__punctuation__(_txt[i]):
                    if i0 is not None:
                        _word = _txt[i0:i].lower()
                        if _word not in _set:
                            highlights[(i0,i)] = _co    
                        i0 = None
                else:
                    if i0 is None:
                        i0 = i
                i += 1
            if i0 is not None:
                _word = _txt[i0:i].lower()
                if _word not in _set:
                    highlights[(i0,i)] = _co
            return highlights

        # Composition
        rttb_main = self.textBlock(text_main, txt_h=main_txt_h, word_wrap=True, w=main_w)
        summary_tiles = []
        for summary_desc in text_summaries:
            _summary = text_summaries[summary_desc]
            rttb_summary = self.textBlock(_summary, txt_h=summary_txt_h, word_wrap=True, w=summary_w)
            summary_tiles.append(f'<svg x="0" y="0" width="{summary_w}" height="{24}">' + \
                                 f'<rect x="0" y="0" width="{summary_w}" height="{24}" fill="#000000" />' + \
                                 self.svgText(summary_desc, 3, 20, txt_h=19, color='#ffffff') + '</svg>')
            summary_tiles.append(rttb_summary.highlights(highlightsForText(_summary, main_words, 'orange'), opacity=opacity))
            summary_tiles.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>') # Spacers

        return self.tile([self.tile(summary_tiles, horz=False)._repr_svg_(),
                          f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>',
                          rttb_main.highlights(highlightsForText(text_main, summary_words, 'yellow'), opacity=opacity)])

    #
    # __textCompareSummaries__sentence_embeddings__()
    #
    def __textCompareSummaries__sentence_embeddings__(self,
                                                      text_main,
                                                      text_summaries,
                                                      embed_fn,
                                                      main_txt_h,
                                                      summary_txt_h,
                                                      spacing,
                                                      opacity,
                                                      w,
                                                      render_histogram=False):
        import umap
        from sklearn.preprocessing import StandardScaler

        # Geometry
        main_w        = summary_w = (w - spacing)/2

        # Colors
        _colors   = self.co_mgr.brewerColors('qualitative', 12) # max available qualitative colors
        _colors_i = 0

        # Text Blocks
        main_rttb     = self.textBlock(text_main, txt_h=main_txt_h, w=main_w, word_wrap=True)
        summary_rttbs = []
        summary_rttb_to_desc = {}

        for _summary_desc in text_summaries:
            _summary = text_summaries[_summary_desc]
            _rttb    = self.textBlock(_summary, txt_h=summary_txt_h, w=summary_w, word_wrap=True)
            summary_rttbs.append(_rttb)
            summary_rttb_to_desc[_rttb] = _summary_desc
        
        # Embeddings
        main_sentences            = self.textExtractSentences(text_main)
        main_sentences_only       = list(list(zip(*main_sentences))[0])
        main_sentences_embeddings = embed_fn(main_sentences_only)
        main_sentence_colors      = {} # [sentence_index] = hex-color-string

        for_umap_source           = []
        for_umap_num              = []
        for_umap_sentence_num     = []
        for_umap_sentence         = []

        for_umap_embeddings       = []
        for_umap_embeddings.extend(main_sentences_embeddings)

        for i in range(0,len(main_sentences_only)):
            for_umap_source.      append('main')
            for_umap_num.         append(0)
            for_umap_sentence_num.append(i)
            for_umap_sentence.    append(main_sentences_only[i])

        # For every summary supplied...
        summary_dots_lu,summary_highlights,min_dot,max_dot,summary_num = {},{},None,None,0
        summary_highlights_lu = {} # [summary][summary_sentence_index] = best_found_main_sentence_index
        summary_num_to_desc   = {}
        for _summary_desc in text_summaries:
            _summary                      = text_summaries[_summary_desc]
            _summary_sentences            = self.textExtractSentences(_summary)
            _summary_sentences_only       = list(list(zip(*_summary_sentences))[0])
            _summary_sentences_embeddings = embed_fn(_summary_sentences_only)

            for_umap_embeddings.extend(_summary_sentences_embeddings)
            for i in range(0,len(_summary_sentences_only)):
                for_umap_source.       append('summary')
                for_umap_num.          append(summary_num)
                for_umap_sentence_num. append(i)
                for_umap_sentence.     append(_summary_sentences_only[i])

            summary_dots = []
            summary_highlights[_summary]    = {}
            summary_highlights_lu[_summary] = {}
            # For every sentence in this summary...
            for i in range(0,len(_summary_sentences)):
                _embedding = _summary_sentences_embeddings[i]
                dots = []
                best_dot,best_dot_main_sentence_index = None,None

                # Loop over the main sentence embeddings -- record both all the dot products as well as the best main sentence match
                for j in range(0,len(main_sentences)):
                    # Get the main sentence embedding
                    _main_embedding = main_sentences_embeddings[j]
                    # Compute the dot product between the main sentence and this specific summaries sentence
                    _dot = float(np.tensordot(_embedding, _main_embedding, axes=1)) # Works with Google's Universal Sentence Embedder...
                    if min_dot is None or min_dot > _dot:
                        min_dot = _dot
                    if max_dot is None or max_dot < _dot:
                        max_dot = _dot                        
                    dots.append(_dot)

                    # Record the best dot found so far (vs the main sentences)
                    if best_dot is None or best_dot < _dot: # Looking for the largest based on some testing...
                        best_dot                     = _dot
                        best_dot_main_sentence_index = j

                summary_dots.append(dots)

                # Try to highlight (if we found something -- how could we not? ... and if we have any colors left)
                if best_dot_main_sentence_index is not None:
                    summary_highlights_lu[_summary][i] = best_dot_main_sentence_index
                    if   best_dot_main_sentence_index in main_sentence_colors.keys(): # Already found!
                        beg_end = (_summary_sentences[i][-2], _summary_sentences[i][-1])
                        summary_highlights[_summary][beg_end] = main_sentence_colors[best_dot_main_sentence_index]
                    elif _colors_i < len(_colors):                                    # Still Have Colors Left!
                        main_sentence_colors[best_dot_main_sentence_index] = _colors[_colors_i]
                        _colors_i += 1
                        beg_end = (_summary_sentences[i][-2], _summary_sentences[i][-1])
                        summary_highlights[_summary][beg_end] = main_sentence_colors[best_dot_main_sentence_index]
                    else:                                                             # No Colors Left :(
                        pass

            summary_dots_lu     [_summary]      =  summary_dots
            summary_num_to_desc [summary_num]   =  _summary_desc
            summary_num                        +=  1

    
        # Create the main highlights
        main_highlights = {}
        for i in main_sentence_colors:
            _tup = main_sentences[i]
            main_highlights[(_tup[-2],_tup[-1])] = main_sentence_colors[i]

        # Renderings & Compositions
        summary_tiles = []
        for _rttb in summary_rttbs:
            _desc = summary_rttb_to_desc[_rttb]
            summary_tiles.append(f'<svg x="0" y="0" width="{summary_w}" height="{24}">' + \
                                 f'<rect x="0" y="0" width="{summary_w}" height="{24}" fill="#000000" />' + \
                                 self.svgText(_desc, 3, 20, txt_h=19, color='#ffffff') + '</svg>')
            summary_tiles.append(_rttb.highlights(summary_highlights[_rttb.txt], opacity=opacity))
            summary_tiles.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>') # Spacers
            summary_tiles.append(self.__textDotProductHeatMap__(summary_dots_lu[_rttb.txt], min_dot, max_dot, 
                                                                summary_highlights_lu[_rttb.txt], main_sentence_colors))
            summary_tiles.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>') # Spacers
            if render_histogram:
                summary_tiles.append(self.__textDotProductHistogram__(summary_dots_lu[_rttb.txt],
                                                                    summary_highlights_lu[_rttb.txt], main_sentence_colors, 
                                                                    summary_w))
                summary_tiles.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>') # Spacers

        # Create the UMAP
        umap_reducer               = umap.UMAP()
        scaled_for_umap_embeddings = StandardScaler().fit_transform(for_umap_embeddings) 
        umap_embedding             = umap_reducer.fit_transform(scaled_for_umap_embeddings)
        umap_xs,umap_ys            = [],[]
        for i in range(0,len(umap_embedding)):
            umap_xs.append(umap_embedding[i][0])
            umap_ys.append(umap_embedding[i][1])
        umap_color,umap_size = [],[]
        for i in range(0,len(umap_embedding)):
            sentence_num = for_umap_sentence_num[i]
            # Dot Color...
            if for_umap_source[i] == 'main':
                if sentence_num in main_sentence_colors.keys():
                    umap_color.append(main_sentence_colors[sentence_num])
                else:
                    umap_color.append('#808080')
            else:
                _summary_desc = summary_num_to_desc[for_umap_num[i]]
                _summary      = text_summaries[_summary_desc]
                if for_umap_sentence_num[i] in summary_highlights_lu[_summary].keys():
                    closest_main_i = summary_highlights_lu[_summary][for_umap_sentence_num[i]]
                    if closest_main_i in main_sentence_colors.keys():
                        umap_color.append(main_sentence_colors[closest_main_i])
                    else:
                        umap_color.append('#808080')    
                else:
                    umap_color.append('#808080')
            # Dot Size...
            if   sentence_num == 0:
                umap_size.append(2.0)
            elif sentence_num == 1:
                umap_size.append(1.8)
            elif sentence_num == 2:
                umap_size.append(1.5)
            elif sentence_num == 3:
                umap_size.append(1.3)
            else:
                umap_size.append(1.0)

        df_umap = pd.DataFrame({'sentence':    for_umap_sentence,
                                'setence_num': for_umap_sentence_num,
                                'source_type': for_umap_source,
                                'source_num':  for_umap_num,
                                'color':       umap_color,
                                'size':        umap_size,
                                'x_umap':      umap_xs,
                                'y_umap':      umap_ys})
        
        def _mydotshape_(_df, _k, _x, _y, _local_dot_w, _color, _opacity):
            source_type = _df['source_type'].iloc[0]
            source_num  = _df['source_num'] .iloc[0]
            if source_type == 'main':
                return 'x'
            else:
                if   source_num == 0:
                    return 'square'
                elif source_num == 1:
                    return 'ellipse'
                else:
                    return 'triangle'

        # Skip the umap for now... doesn't add anything...
        # summary_tiles.append(self.xy(df_umap, x_field='x_umap', y_field='y_umap', color_by='color', count_by='size', dot_size='vary', dot_shape=_mydotshape_, draw_labels=False))

        # Compose the summary side
        tile_composition = self.tile(summary_tiles, horz=False)._repr_svg_()

        # Compose the total
        composition = [tile_composition,
                       f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>',
                       main_rttb.highlights(main_highlights, opacity=opacity)]
        
        return self.tile(composition)
    
    #
    # __textDotProductXYDataFrame__
    #
    def __textDotProductHistogram__(self, arr, _sentence_index_to_main_index, _main_index_colors, _w):
        _xs,_ys,_colors,_groups = [],[],[],[]
        for _group in range(0,len(arr)):
            _copy = sorted(np.array(arr[_group]),reverse=True)
            for x in range(0,len(_copy)):
                y = _copy[x]
                _xs.     append(x)
                _ys.     append(y)
                _groups. append(_group)
                _color = '#000000'
                if _group in _sentence_index_to_main_index.keys():
                    main_index = _sentence_index_to_main_index[_group]
                    if main_index in _main_index_colors.keys():
                        _color = _main_index_colors[main_index]
                _colors.append(_color)
        _df = pd.DataFrame({'x':_xs,'y':_ys,'color':_colors,'group':_groups})
        return self.xy(_df, x_field='x', y_field='y', color_by='color', line_groupby_field='group', line_groupby_w=2.0, dot_size=None, w=_w, h=128)

    #
    # _textDotProductHeatMap__():  Make a simplified heatmap
    #
    def __textDotProductHeatMap__(self, arr, _min, _max, _sentence_index_to_main_index, _main_index_colors):
        if _min == _max:
            _max = _min + 1
        x_tiles,y_tiles = len(arr[0]),len(arr)
        tile_w, tile_h  = 12,12
        svg = f'<svg x="0" y="0" width="{x_tiles*tile_w + 3*tile_w}" height="{y_tiles*tile_h}">'
        for y in range(0,len(arr)):
            for x in range(0,len(arr[y])):
                _value = arr[y][x]
                # _color = self.co_mgr.spectrumAbridged(_value, _min, _max)
                _gray    = min(255, int(255 * (_value - _min)/(_max - _min)))
                _color   = f'#{_gray:02x}{_gray:02x}{_gray:02x}'
                svg += f'<rect x="{x*tile_w}" y="{y*tile_h}" width = "{tile_w}" height="{tile_h}" fill="{_color}" />'
            if y in _sentence_index_to_main_index.keys():
                main_index = _sentence_index_to_main_index[y]
                if main_index in _main_index_colors.keys():
                    _color = _main_index_colors[main_index]
                    svg += f'<rect x="{x_tiles*tile_w + tile_w}" y="{y*tile_h}" width = "{2*tile_w}" height="{tile_h}" fill="{_color}" />'
        svg += '</svg>'
        return svg

    #
    # textCreateEmbedder() - Create an embedder
    #
    def textCreateEmbedder(self, desc='google_universal_sentence_embedder'):
        import tensorflow_hub as hub
        module_url = "https://tfhub.dev/google/universal-sentence-encoder/4"
        model = hub.load(module_url)
        return model

    #
    # textCreateRoBertModel() - Create a Bert model
    #
    def textCreateRoBERTaModel(self, 
                               model_name='roberta-base'):
        from transformers import AutoTokenizer, RobertaForMaskedLM
        import torch
        device    = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        tokenizer = AutoTokenizer.     from_pretrained(model_name)
        model     = RobertaForMaskedLM.from_pretrained(model_name)
        model.to(device)                                                                                     # and move our model over to the selected device
        return model,tokenizer,device

    #
    # __textTrainRoBERTatModel__()
    # - Modified from:
    #   https://towardsdatascience.com/transformers-retraining-roberta-base-using-the-roberta-mlm-procedure-7422160d5764
    #
    def __textTrainRoBERTaModel__(self, text_main, epochs=100):
        from transformers import RobertaTokenizer, RobertaForMaskedLM
        import torch
        tokenizer = RobertaTokenizer.   from_pretrained('roberta-base')
        model     = RobertaForMaskedLM. from_pretrained('roberta-base')
        class GeneratorFromListOfStrings(object):
            def __init__(self,txts,tokenizer):
                self.txts      = txts
                self.tokenizer = tokenizer
                self.i    = 0
            def __getitem__(self, index):
                return {'input_ids': self.tokenizer.encode(self.txts[index], return_tensors='pt')[0]}
            def __len__(self):
                return len(self.txts)
            def __iter__(self):
                return self
            def __next__(self):        
                if self.i < len(self.txts):            
                    _result_ = {'input_ids': self.tokenizer.encode(self.txts[self.i], return_tensors='pt')[0]}
                    self.i += 1
                    return _result_
                else:
                    raise StopIteration()
            def __call__(self):
                self.i = 0
                return self
        gflos = GeneratorFromListOfStrings(text_main.split('\n'), tokenizer)
        from datasets import Dataset
        ds_from_gflos = Dataset.from_generator(gflos)
        from transformers import DataCollatorForLanguageModeling
        data_collator = DataCollatorForLanguageModeling(tokenizer=tokenizer, mlm=True, mlm_probability=0.15)
        from transformers import Trainer, TrainingArguments
        training_args = TrainingArguments(
            output_dir="./roberta-retrained",
            overwrite_output_dir=True,
            num_train_epochs=epochs,
            per_device_train_batch_size=48,
            save_steps=500,
            save_total_limit=2,
            seed=1
        )
        trainer = Trainer(
            model=model,
            args=training_args,
            data_collator=data_collator,
            train_dataset=ds_from_gflos
        )
        trainer.train()
        model.eval()
        device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        return model, tokenizer, device

    #
    # __textRoBERTsStats__()
    # ... evolving method to determine probabilities/rankings for a specific word in a document.
    #
    def __textRoBERTaStats__(self, sentence, model, tokenizer, device, k=40, bins=20):
        import torch
        import tensorflow as tf

        results   = []
        _inputs   = tokenizer(sentence, return_tensors="pt")
        i0        = 0
        for token_i in range(1,len(_inputs['input_ids'][0])-2):
            _token                  = tokenizer.decode(_inputs['input_ids'][0][token_i])
            i0 = sentence.index(_token, i0)
            _token_stripped_lowered = _token.strip().lower()
            _inputs_w_mask                          = tokenizer(sentence, return_tensors="pt")
            _inputs_w_mask['input_ids'][0][token_i] = tokenizer.encode('<mask>')[1]
            _inputs_w_mask.to(device)
            with torch.no_grad():
                _output = model(**_inputs_w_mask)
                _logits = _output.logits
            predicted_token_id = _logits[0, token_i].argmax(axis=-1)
            _predicted = tokenizer.decode(predicted_token_id)
            top_k_indices = tf.math.top_k(_logits.cpu().detach().numpy(), k).indices[0].numpy()
            i,ith = 0,None
            for x in top_k_indices[token_i]:
                if tokenizer.decode(x).strip().lower() == _token_stripped_lowered:
                    if ith is None:
                        ith = i
                i += 1
            labels = tokenizer(sentence, return_tensors="pt")["input_ids"]
            word_score = float(_output[0][0][token_i][labels[0][token_i]])
            # From https://github.com/pytorch/pytorch/issues/69519
            def histogram(xs, bins):
                # Like torch.histogram, but works with cuda
                min, max = xs.min(), xs.max()
                counts     = torch.histc(xs, bins, min=min, max=max)
                boundaries = torch.linspace(min, max, bins + 1)
                return counts, boundaries
            _counts,_boundaries = histogram(_output[0][0][token_i], bins=bins)
            results.append({'token':_token, 'i0':i0, 'i1':i0+len(_token), 'predicted':_predicted, 'score':word_score, 
                            'ith':ith, 'counts':_counts.cpu(), 'boundaries':_boundaries.cpu()})
        return results
            
    #
    # textCreateBertModel() - Create a Bert model
    #
    # model_name
    # - 'bert-base-cased'   <== Default
    # - 'bert-large-cased'
    #
    def textCreateBertModel(self, 
                            model_name='bert-base-cased'):
        from transformers import BertTokenizer, BertForMaskedLM
        import torch
        device    = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        tokenizer = BertTokenizer.  from_pretrained(model_name)
        model     = BertForMaskedLM.from_pretrained(model_name)
        model.to(device)                                                                                     # and move our model over to the selected device
        return model,tokenizer,device

    #
    # __textTrainBertModel__()
    # ... used the"lm_labels" variable ... seems to be deprecated
    #
    def __OLD_textTrainBertModel__(self, text_main, mask_perc=0.75, epochs=100):
        from transformers import BertTokenizer, BertForMaskedLM, TFBertForMaskedLM, AdamW       
        import torch 

        # From the throwaway file "bert_mlm_example.ipynb"
        #
        # Modified From https://towardsdatascience.com/masked-language-modelling-with-bert-7d49793e5d2c
        #
        device    = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        tokenizer = BertTokenizer.  from_pretrained('bert-base-cased')
        model     = BertForMaskedLM.from_pretrained('bert-base-cased')
        #tokenizer = BertTokenizer.  from_pretrained('bert-large-cased')
        #model     = BertForMaskedLM.from_pretrained('bert-large-cased')        
        model.train()                                                                                        # activate training mode
        model.to(device)                                                                                     # and move our model over to the selected device
        optim = None
        _parts = text_main.split('\n')
        for epoch in range(epochs):
            for _part in _parts:
                text = (_part)
                as_encoded = tokenizer.encode(text)
                inputs = {'input_ids':     torch.Tensor([as_encoded]).long(),
                          'token_type_ids':torch.Tensor([np.zeros(len(as_encoded))]).long(),
                          'attention_mask':torch.Tensor([np.ones (len(as_encoded))]).long()}
                inputs['lm_labels'] = inputs['input_ids'].detach().clone()                                   # labels are just the original text...
                rand      = torch.rand(inputs['input_ids'].shape)                                            # create random array of floats in equal dimension to input_ids
                mask_arr  = (rand < mask_perc) * (inputs['input_ids'] != 101) * (inputs['input_ids'] != 102) # As an example of how to separate out those two token types
                selection = torch.flatten((mask_arr[0]).nonzero()).tolist()                                  # create selection from mask_arr
                inputs['input_ids'][0, selection] = 103                                                      # apply selection index to inputs.input_ids, adding MASK tokens
                input_ids      = inputs['input_ids'].     to(device)                                         # move to gpu
                attention_mask = inputs['attention_mask'].to(device)
                lm_labels      = inputs['lm_labels'].     to(device)    
                outputs = model(input_ids, attention_mask=attention_mask, lm_labels=lm_labels)               # process
                if optim is None:
                    optim = AdamW(model.parameters(), lr=5e-5)                                               # initialize optimizer
                optim.zero_grad()                                                                            # initialize calculated gradients (from prev step)
                input_ids      = inputs['input_ids'].     to(device)                                         # move to gpu
                attention_mask = inputs['attention_mask'].to(device)
                lm_labels      = inputs['lm_labels'].     to(device)    
                outputs        = model(input_ids, attention_mask=attention_mask, lm_labels=lm_labels)        # process
                loss           = outputs[0]                                                                  # extract loss
                loss.backward()                                                                              # calculate loss for every parameter that needs grad update    
                optim.step()                                                                                 # update parameters    
            if (epoch%10) == 0:                                                                              # print updated information
                print(f'Epoch {epoch:3}\t{loss.item()}')
        model.eval()                                                                                         # Take it out of training mode
        return model,tokenizer,device

    #
    # __textTrainBertModel__()
    #
    def __textTrainBertModel__(self, text_main, mask_perc=0.75, epochs=100):
        from transformers import BertTokenizer, BertForMaskedLM, TFBertForMaskedLM, AdamW
        import torch

        # From the throwaway file "bert_mlm_example.ipynb"
        #
        # Modified From https://towardsdatascience.com/masked-language-modelling-with-bert-7d49793e5d2c
        #
        device    = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        tokenizer = BertTokenizer.  from_pretrained('bert-base-cased')
        model     = BertForMaskedLM.from_pretrained('bert-base-cased')
        #tokenizer = BertTokenizer.  from_pretrained('bert-large-cased')
        #model     = BertForMaskedLM.from_pretrained('bert-large-cased')        
        model.train()                                                                                        # activate training mode
        model.to(device)                                                                                     # and move our model over to the selected device
        optim = None
        _parts = text_main.split('\n')
        for epoch in range(epochs):
            for _part in _parts:
                text = (_part)
                as_encoded = tokenizer.encode(text)
                inputs = {'input_ids':     torch.Tensor([as_encoded]).long(),
                          'token_type_ids':torch.Tensor([np.zeros(len(as_encoded))]).long(),
                          'attention_mask':torch.Tensor([np.ones (len(as_encoded))]).long()}
                inputs['labels'] = inputs['input_ids'].detach().clone()                                      # labels are just the original text...
                rand      = torch.rand(inputs['input_ids'].shape)                                            # create random array of floats in equal dimension to input_ids
                mask_arr  = (rand < mask_perc) * (inputs['input_ids'] != 101) * (inputs['input_ids'] != 102) # As an example of how to separate out those two token types
                selection = torch.flatten((mask_arr[0]).nonzero()).tolist()                                  # create selection from mask_arr
                inputs['input_ids'][0, selection] = 103                                                      # apply selection index to inputs.input_ids, adding MASK tokens
                input_ids      = inputs['input_ids'].     to(device)                                         # move to gpu
                attention_mask = inputs['attention_mask'].to(device)
                labels         = inputs['labels'].        to(device)    
                outputs = model(input_ids, attention_mask=attention_mask, labels=labels)                     # process
                if optim is None:
                    optim = AdamW(model.parameters(), lr=5e-5)                                               # initialize optimizer
                optim.zero_grad()                                                                            # initialize calculated gradients (from prev step)
                input_ids      = inputs['input_ids'].     to(device)                                         # move to gpu
                attention_mask = inputs['attention_mask'].to(device)
                labels         = inputs['labels'].        to(device)    
                outputs        = model(input_ids, attention_mask=attention_mask, labels=labels)              # process
                loss           = outputs[0]                                                                  # extract loss
                loss.backward()                                                                              # calculate loss for every parameter that needs grad update    
                optim.step()                                                                                 # update parameters    
            if (epoch%10) == 0:                                                                              # print updated information
                print(f'Epoch {epoch:3}\t{loss.item()}')
        model.eval()                                                                                         # Take it out of training mode
        return model,tokenizer,device

    #
    # __textBertTopKPredictions__():  Return the top k predictions for the [MASK] word.
    #
    # From https://mattmckenna.io/bert-off-the-shelf/#:~:text=BERT%20works%20by%20masking%20certain,ate%20the%20%5BMASK%5D%E2%80%9D.
    # - with a lot of modifications ...  for the GPU version...
    def __textBertTopKPredictions__(self, input_string, k, model, tokenizer, device):
        import torch
        import tensorflow as tf
        tokenized_inputs = tokenizer.encode(input_string)
        outputs = model(torch.Tensor([tokenized_inputs]).long().to(device))
        top_k_indices = tf.math.top_k(outputs[0].cpu().detach().numpy(), k).indices[0].numpy()
        for i in range(len(tokenized_inputs)):
            if tokenized_inputs[i] == tokenizer.encode(tokenizer.mask_token)[1]:
                mask_i = i
        return tokenizer.decode(top_k_indices[mask_i])

    #
    # __textBertWordProbabilities__():  Return the associated probabilities for the words.
    #
    # From https://mattmckenna.io/bert-off-the-shelf/#:~:text=BERT%20works%20by%20masking%20certain,ate%20the%20%5BMASK%5D%E2%80%9D.
    #
    def __textBertWordProbabilities__(self, input_string, model, tokenizer, device):
        import torch
        tokenized_inputs = tokenizer.encode(input_string)
        outputs = model(torch.Tensor([tokenized_inputs]).long().to(device))
        predictions = outputs[0]
        predicted_indices = torch.argmax(predictions[0,:],dim=-1)
        predicted_tokens  = tokenizer.convert_ids_to_tokens(predicted_indices.tolist())
        probs = torch.nn.functional.softmax(predictions, dim=-1)
        predicted_token_probs = probs[0,torch.arange(predictions.shape[1]),predicted_indices].cpu()
        return predicted_tokens[1:-1], predicted_token_probs[1:-1], tokenizer.tokenize(input_string), predictions.cpu()[0,1:-1], tokenized_inputs[1:-1]

    #
    # __textCompareSummaries__bert_top_n__():  Compare via top-n bert placements
    #
    def __textCompareSummaries__bert_top_n__(self,
                                             text_main, 
                                             text_summaries,
                                             methodology,
                                             model,
                                             tokenizer,
                                             device,
                                             main_txt_h, 
                                             summary_txt_h, 
                                             spacing, 
                                             opacity, 
                                             w):
        # Geometry & Parameter Evaluation
        main_w = summary_w = (w - spacing)/2
        if type(text_summaries) == str:
            text_summaries = {'Default':text_summaries}

        # Create the model if necessary
        if model is None:
            model,tokenizer,device = self.__textTrainBertModel__(text_main)

        # Put the two last functions together for input highlights text input...
        def highlightsForText(_txt):
            highlights = {}
            _parts   = _txt.split('\n')
            accum_i = 0
            for _part in _parts:
                parts_i, global_i = 0, accum_i
                while parts_i < len(_part):
                    while parts_i < len(_part) and (self.__whitespace__(_part[parts_i]) or self.__punctuation__(_part[parts_i])):
                        parts_i  += 1
                        global_i += 1
                    if parts_i < len(_part):
                        parts_i0,global_i0 = parts_i,global_i
                        while parts_i < len(_part) and self.__whitespace__(_part[parts_i]) == False and self.__punctuation__(_part[parts_i]) == False:
                            parts_i  += 1
                            global_i += 1
                        actual        = _part[parts_i0:parts_i]
                        masked        = _part[:parts_i0] + '[MASK]' + _part[parts_i:]
                        guesses       = self.__textBertTopKPredictions__(masked, 20, model, tokenizer, device)
                        guesses_split = guesses.split(' ')
                        guesses_i     = 0
                        while guesses_i < len(guesses_split) and guesses_split[guesses_i] != actual:
                            guesses_i += 1
                        if guesses_i <= 1:
                            _co = None

                        if   guesses_i <= 1:
                            _co = None
                        elif guesses_i <= 5:
                            _co = '#909090'
                        elif guesses_i <= 10:
                            _co = 'yellow'
                        elif guesses_i <  20:
                            _co = 'orange'
                        else:
                            _co = 'red'
                        if _co is not None:
                            i0_to_i1 = (global_i0, global_i)
                            highlights[i0_to_i1] = _co
                accum_i += len(_part)+1
            return highlights

        if methodology == 'bert_top_n':
            rttb_main = self.textBlock(text_main, txt_h=main_txt_h, word_wrap=True, w=main_w)
            summary_tiles = []
            for summary_desc in text_summaries:
                _summary = text_summaries[summary_desc]
                rttb_summary = self.textBlock(_summary, txt_h=summary_txt_h, word_wrap=True, w=summary_w)
                summary_tiles.append(f'<svg x="0" y="0" width="{summary_w}" height="{24}">' + \
                                    f'<rect x="0" y="0" width="{summary_w}" height="{24}" fill="#000000" />' + \
                                    self.svgText(summary_desc, 3, 20, txt_h=19, color='#ffffff') + '</svg>')
                summary_tiles.append(rttb_summary.highlights(highlightsForText(_summary), opacity=opacity))
                summary_tiles.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>') # Spacers

            return self.tile([self.tile(summary_tiles, horz=False)._repr_svg_(),
                            f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>',
                            rttb_main.highlights(highlightsForText(text_main), opacity=opacity)])

        def tokenSpans(orig, as_tokens):
            spans = []
            i,token_i = 0,0
            while token_i < len(as_tokens):
                token = as_tokens[token_i]
                if token.startswith('##'):
                    token = token[2:]
                i = orig.index(token,i)
                spans.append((i,i+len(token)))
                i += len(token)
                token_i += 1
            return spans

        def probabilityHighlights(orig):
            highlights = {}
            _parts,_parts_i = orig.split('\n'),0
            for _part in _parts:
                pred_tokens, pred_probs, as_tokens, preds, token_inputs = self.__textBertWordProbabilities__(_part, model, tokenizer, device)
                spans = tokenSpans(_part, as_tokens)
                for i in range(len(pred_probs)):
                    _span = (spans[i][0]+_parts_i,spans[i][1]+_parts_i)
                    if   pred_probs[i] > 0.95:
                        pass
                    elif pred_probs[i] > 0.8:
                        highlights[_span] = '#909090'
                    elif pred_probs[i] > 0.6:
                        highlights[_span] = 'yellow'
                    elif pred_probs[i] > 0.4:
                        highlights[_span] = 'orange'
                    else:
                        highlights[_span] = 'red'
                _parts_i += len(_part) + 1
            return highlights

        rttb_main = self.textBlock(text_main, txt_h=main_txt_h, word_wrap=True, w=main_w)
        summary_tiles = []
        for summary_desc in text_summaries:
            _summary = text_summaries[summary_desc]
            rttb_summary = self.textBlock(_summary, txt_h=summary_txt_h, word_wrap=True, w=summary_w)
            summary_tiles.append(f'<svg x="0" y="0" width="{summary_w}" height="{24}">' + \
                                 f'<rect x="0" y="0" width="{summary_w}" height="{24}" fill="#000000" />' + \
                                 self.svgText(summary_desc, 3, 20, txt_h=19, color='#ffffff') + '</svg>')
            summary_tiles.append(rttb_summary.highlights(probabilityHighlights(_summary), opacity=opacity))
            summary_tiles.append(f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>') # Spacers

        return self.tile([self.tile(summary_tiles, horz=False)._repr_svg_(),
                          f'<svg x="0" y="0" width="{spacing}" height="{spacing}"> </svg>',
                          rttb_main.highlights(probabilityHighlights(text_main), opacity=opacity)])

#
# RTTextBlock - instance of rendered text block
#
class RTTextBlock(object):
    #
    # Constructor
    #
    def __init__(self,
                 rt_self,           # Reference to parent class instance
                 txt,               # Original text string
                 txt_h,             # Text height in pixels
                 line_space_px,     # Pixel space between paragraphs
                 word_wrap,         # Word wrap flag
                 w,                 # Width of SVG results
                 x_ins,             # x insert left & right
                 y_ins,             # y insert top & bottom
                 svg,               # rendered svg (w/out svg begin/end wrapper)
                 bounds,            # Four tuple of x,y,w,h
                 geom_to_word,      # Shapely polygon to word
                 orig_to_xy):       # Original text index to xy-tuple
        self.rt_self        = rt_self
        self.txt            = txt
        self.txt_h          = txt_h
        self.line_space_px  = line_space_px
        self.word_wrap      = word_wrap
        self.w              = w
        self.x_ins          = x_ins
        self.y_ins          = y_ins
        self.svg            = svg
        self.bounds         = bounds
        self.geom_to_word   = geom_to_word
        self.orig_to_xy     = orig_to_xy
        
    #
    # spanGeometry() - return a polygon that covers a specified text span.
    #
    def spanGeometry(self, i, j):
        last_c = ' '
        if len(self.txt) > 0:
            last_c = self.txt[-1]

        if i >= len(self.txt):
            xy0    = self.orig_to_xy[len(self.txt)-1]
            xy0    = (xy0[0] + self.rt_self.textLength(last_c,self.txt_h),xy0[1])
        else:
            xy0    = self.orig_to_xy[i]
        if j >= len(self.txt):
            xy1    = self.orig_to_xy[len(self.txt)-1]
            xy1    = (xy1[0] + self.rt_self.textLength(last_c,self.txt_h),xy1[1])
        else:
            xy1    = self.orig_to_xy[j]

        _dn = 4 # downward shift...
        if     xy0[1]                                    == xy1[1]: # On same line...
            return Polygon([[xy0[0],xy0[1]+_dn],
                            [xy1[0],xy1[1]+_dn],
                            [xy1[0],xy1[1]-self.txt_h+_dn],
                            [xy0[0],xy0[1]-self.txt_h+_dn]
                            ])
        elif  (xy0[1] + self.txt_h + self.line_space_px) == xy1[1] and (xy1[0] < xy0[0]):
            _poly0 = Polygon([[xy0[0],              xy0[1]+_dn],
                              [self.w - self.x_ins, xy0[1]+_dn],
                              [self.w - self.x_ins, xy0[1]-self.txt_h+_dn],
                              [xy0[0],              xy0[1]-self.txt_h+_dn]])
            _poly1 = Polygon([[xy1[0],              xy1[1]+_dn],
                              [xy1[0],              xy1[1]-self.txt_h+_dn],
                              [self.x_ins,          xy1[1]-self.txt_h+_dn],
                              [self.x_ins,          xy1[1]+_dn]])
            return MultiPolygon([_poly0,_poly1])
        else: # Multiple lines...
            return Polygon([[xy0[0],              xy0[1]+_dn],
                            [self.x_ins,          xy0[1]+_dn],
                            [self.x_ins,          xy1[1]+_dn],
                            [xy1[0],              xy1[1]+_dn],
                            [xy1[0],              xy1[1]-self.txt_h+_dn],
                            [self.w - self.x_ins, xy1[1]-self.txt_h+_dn],
                            [self.w - self.x_ins, xy0[1]-self.txt_h+_dn],
                            [xy0[0],              xy0[1]-self.txt_h+_dn]
                            ])

    #
    # pixelRepr() - return a pixel level representation of the document
    # - see the highlights() function below for regex formatting
    #
    def pixelRepr(self, lu, w=64):
        bounds_x,bounds_y,bounds_w,bounds_h = self.bounds
        scale = w / bounds_w
        h = w * bounds_h / bounds_w
        _co  = self.rt_self.co_mgr.getTVColor('background','default')
        svg  = f'<svg x="0" y="0" width="{w}" height="{h}">'
        svg += f'<rect x="0" y="0" width="{w}" height="{h}" fill="{_co}" />'
        for k in lu.keys():
            if   type(k) == tuple:
                _poly        = self.spanGeometry(k[0],k[1])
                _poly_scaled = affinity.scale(_poly,xfact=scale,yfact=scale,origin=(0,0,0))
                _co          = self.rt_self.co_mgr.getColor(lu[k])
                svg += f'<path d="{self.rt_self.shapelyPolygonToSVGPathDescription(_poly_scaled)}" fill="{_co}" />'
            elif type(k) == str:
                re_match = re.findall(k,self.txt)
                if re_match is not None and len(re_match) > 0:
                    i = 0
                    for _match in re_match:
                        if type(_match) == tuple:
                            _match = _match[0]
                        i = self.txt.index(_match,i)
                        j = i + len(_match)
                        _poly        = self.spanGeometry(i,j)
                        _poly_scaled = affinity.scale(_poly,xfact=scale,yfact=scale,origin=(0,0,0))
                        _co          = self.rt_self.co_mgr.getColor(lu[k])
                        svg += f'<path d="{self.rt_self.shapelyPolygonToSVGPathDescription(_poly_scaled)}" fill="{_co}" />'
                        i += len(_match)
            else:
                raise Exception(f'RTTextBlock.pixelRepr() -- unknown key in lookups {k}')
        svg +=  '</svg>'
        return svg

    #
    # highlights() - highlight user-specified text.
    # - lu: either a [1] (i,j) tuple or [2] a regex string to either a [A] seven-character hex color string or a [B] string color
    #   - lu[(0,10)]            = '#ff0000'
    #   - lu['regex substring'] = '#000000' -- this needs to be grouped properly -- for example r'(([Mm]atch)(es){0,1})
    #   - lu['many']            = 'whatever' # any 'many' substrings will get colored with 'whatever' color lookup
    #
    def highlightsOverlay(self, lu, opacity=0.4):
        svg_underlay = ''
        for k in lu:
            _co = lu[k]
            if _co.startswith('#') == False or len(_co) != 7: # If it's not a hex hash color string... then look it up...
                _co = self.rt_self.co_mgr.getColor(_co)
            if   type(k) == tuple:
                _poly = self.spanGeometry(k[0],k[1])
                svg_underlay += f'<path d="{self.rt_self.shapelyPolygonToSVGPathDescription(_poly)}" fill="{_co}" fill-opacity="{opacity}" />'
            elif type(k) == str:
                re_match = re.findall(k,self.txt)
                if re_match is not None and len(re_match) > 0:
                    i = 0
                    for _match in re_match:
                        if type(_match) == tuple:
                            _match = _match[0]
                        i = self.txt.index(_match,i)
                        j = i + len(_match)
                        _poly = self.spanGeometry(i,j)
                        svg_underlay += f'<path d="{self.rt_self.shapelyPolygonToSVGPathDescription(_poly)}" fill="{_co}" fill-opacity="{opacity}" />'
                        i += len(_match)
            else:
                raise Exception(f'RTTextBlock.highlights() - do not understand key value type {type(k)}')

        x,y,w,h = self.bounds
        return f'<svg x="0" y="0" width="{w}" height="{h}">' + svg_underlay + '</svg>'

    def highlights(self,lu,opacity=0.4):
        return self.wrap(self.background() + self.unwrappedText() + self.highlightsOverlay(lu, opacity))

    #
    # __OLD_underlineSpan__() - internal primitive for underlining.
    # ... this one performed it on a per word basis with gaps where there were spaces... and didn't look right...
    #
    def __OLD_underlineSpan__(self, i0, i1, _co=None, _stroke_w=None, y_offset=0, strikethrough=False):
        y_adj = y_offset
        if strikethrough:
            y_adj = -self.txt_h/2
        if _co is None:
            _co = self.rt_self.co_mgr.getTVColor('data','default')
        if _stroke_w is None:
            _stroke_w = min(0.5 + self.txt_h/14, 2.5)
        my_svg = ''
        x0, x1, y = None,None,None
        for i in range(i0,i1):
            c, xy = self.txt[i], self.orig_to_xy[i]
            if self.rt_self.__whitespace__(c) or self.rt_self.__punctuation__(c):
                if x0 is not None:
                    my_svg += f'<line x1="{x0}" y1="{y + 1 + _stroke_w + y_adj}" x2="{x1}" y2="{y + 1 + _stroke_w  + y_adj}" stroke="{_co}" stroke-width="{_stroke_w}" />'
                    x0, x1, y = None,None,None
            elif x0 is not None and y != xy[1]:
                my_svg += f'<line x1="{x0}" y1="{y + 1 + _stroke_w  + y_adj}" x2="{x1}" y2="{y + 1 + _stroke_w  + y_adj}" stroke="{_co}"  stroke-width="{_stroke_w}" />'
                x0, x1, y = None,None,None
            elif x0 is None:
                x0, x1, y = xy[0], xy[0] + self.rt_self.textLength(c, self.txt_h), xy[1]
            else:
                x1 = xy[0] + self.rt_self.textLength(c, self.txt_h)
        if x0 is not None:
            my_svg += f'<line x1="{x0}" y1="{y + 1 + _stroke_w + y_adj}" x2="{x1}" y2="{y + 1 + _stroke_w + y_adj}" stroke="{_co}"  stroke-width="{_stroke_w}" />'
        return my_svg

    #
    # __underlineSpan__() - internal primitive for underlining.
    #
    def __underlineSpan__(self, i0, i1, _co=None, _stroke_w=None, y_offset=0, strikethrough=False):
        y_adj = y_offset
        if strikethrough:
            y_adj = -self.txt_h/2
        if _co is None:
            _co = self.rt_self.co_mgr.getTVColor('data','default')
        if _stroke_w is None:
            _stroke_w = min(0.5 + self.txt_h/14, 2.5)
        my_svg = ''
        i,x0,x1,last_y = i0,None,None,None
        while i < i1:
            c,xy = self.txt[i],self.orig_to_xy[i]
            if   last_y is None:
                x0,x1,last_y = xy[0],xy[0]+self.rt_self.textLength(c, self.txt_h),xy[1]
            elif last_y == xy[1]:
                x1 = xy[0]+self.rt_self.textLength(c, self.txt_h)
            else:
                my_svg += f'<line x1="{x0}" y1="{last_y + 1 + _stroke_w  + y_adj}" x2="{x1}" y2="{last_y + 1 + _stroke_w  + y_adj}" stroke="{_co}"  stroke-width="{_stroke_w}" />'
                x0,x1,last_y = xy[0],xy[0]+self.rt_self.textLength(c, self.txt_h),xy[1]
            i += 1
        if x0 is not None:
            my_svg += f'<line x1="{x0}" y1="{last_y + 1 + _stroke_w  + y_adj}" x2="{x1}" y2="{last_y + 1 + _stroke_w  + y_adj}" stroke="{_co}"  stroke-width="{_stroke_w}" />'
            x0,x1,last_y = None,None,None
        return my_svg

    #
    # underlines() - same format as above...
    #
    def underlinesOverlay(self, lu, strikethrough=False, y_offset=0, underline_stroke_w=None):
        svg_underlay = ''
        for k in lu:
            _co = lu[k]
            if _co is None:
                _co = self.rt_self.co_mgr.getTVColor('data','default')
            if _co.startswith('#') == False or len(_co) != 7: # If it's not a hex hash color string... then look it up...
                _co = self.rt_self.co_mgr.getColor(_co)
            if   type(k) == tuple:
                svg_underlay += self.__underlineSpan__(k[0], k[1], _co=_co, strikethrough=strikethrough, y_offset=y_offset, _stroke_w=underline_stroke_w)
            elif type(k) == str:
                i = 0
                re_match = re.findall(k,self.txt)
                if re_match is not None:
                    i = 0
                    for _match in re_match:
                        i = self.txt.index(k,i)
                        j = i + len(_match)
                        svg_underlay += self.__underlineSpan__(i, j, _co=_co, strikethrough=strikethrough, y_offset=y_offset, _stroke_w=underline_stroke_w)
                        i += len(_match)
            else:
                raise Exception(f'RTTextBlock.highlights() - do not understand key value type {type(k)}')
        x,y,w,h = self.bounds
        return f'<svg x="0" y="0" width="{w}" height="{h}">' + svg_underlay + '</svg>'

    def underlines(self, lu, strikethrough=False, y_offset=0, underline_stroke_w=2):
        return self.wrap(self.background() + self.unwrappedText() + self.underlinesOverlay(lu, strikethrough, y_offset=y_offset, underline_stroke_w=underline_stroke_w))

    #
    # unwrappedSVG() - return the unwrapped version of the SVG.
    #
    def unwrappedText(self):
        return self.svg

    #
    # wrap() - wrap into an SVG frame of the proper size
    #
    def wrap(self, to_wrap):
        x,y,w,h = self.bounds        
        return f'<svg x="0" y="0" width="{w}" height="{h}">' + to_wrap + '</svg>'

    #
    # background(self): return the background
    #
    def background(self):
        x,y,w,h = self.bounds
        _co     = self.rt_self.co_mgr.getTVColor('background','default')
        return f'<svg x="0" y="0" width="{w}" height="{h}">' + \
               f'<rect x="0" y="0" width="{w}" height="{h}" fill="{_co}" />' + \
                '</svg>'

    #
    # SVG Representation -- adds the svg begin/end markup...
    #
    def _repr_svg_(self):
        x,y,w,h = self.bounds
        _co     = self.rt_self.co_mgr.getTVColor('background','default')
        return f'<svg x="0" y="0" width="{w}" height="{h}">' + \
               f'<rect x="0" y="0" width="{w}" height="{h}" fill="{_co}" />' + \
               self.svg + \
               '</svg>'
    
    #
    # Debugging Original Indices
    #
    def __debug_svgOfOverlayOriginalIndices__(self):
        svg_overlay = ''
        _co = self.rt_self.co_mgr.getTVColor('data','default')
        for i in self.orig_to_xy:
            x,y = self.orig_to_xy[i]
            svg_overlay += f'<line x1="{x}" y1="{y}" x2="{x}" y2="{y-self.txt_h}" stroke="{_co}" />'

        x,y,w,h = self.bounds
        _co     = self.rt_self.co_mgr.getTVColor('background','default')
        return f'<svg width="{w}" height="{h}">' + \
               f'<rect x="0" y="0" width="{w}" height="{h}" fill="{_co}" />' + \
               self.svg + \
               svg_overlay + \
               '</svg>'
    
    #
    # Debugging Word Geometries
    #
    def __debug_svgOfWordColors__(self):
        svg_underlay = ''

        for _poly in self.geom_to_word:
            _word = self.geom_to_word[_poly]
            _co   = self.rt_self.co_mgr.getColor(_word)
            svg_underlay += f'<path d="{self.rt_self.shapelyPolygonToSVGPathDescription(_poly)}" fill="{_co}" />'

        x,y,w,h = self.bounds
        _co     = self.rt_self.co_mgr.getTVColor('background','default')
        return f'<svg width="{w}" height="{h}">' + \
               f'<rect x="0" y="0" width="{w}" height="{h}" fill="{_co}" />' + \
               svg_underlay + \
               self.svg     + \
               '</svg>'

    #
    # Create a positional dataframe of words / sentences / paragraphs.
    #
    def positionalDataFrame(self):
        _txt = []
        _typ = []
        _num = [] # number within that type
        _beg = []
        _end = []
        # Paragraphs
        _parts = self.txt.split('\n')
        i,para_num = 0,0
        for _part in _parts:
            if len(_part) > 0:
                _txt.append(_part)
                _typ.append('para')
                _num.append(para_num)
                _beg.append(i)
                _end.append(i+len(_part))
                para_num += 1
                i += len(_part)
            i += 1 # for the '\n'

        # Sentences
        _sents = self.rt_self.textExtractSentences(self.txt)
        i,sent_num = 0,0
        for _sent in _sents:
            _txt.append(_sent[0])
            _typ.append('sent')
            _num.append(sent_num)
            _beg.append(_sent[-2])
            _end.append(_sent[-1])
            sent_num += 1
            i += len(_sent)

        # Words
        _word = ''
        i,i0,word_num = 0,-1,0
        while i < len(self.txt):
            c = self.txt[i]
            if self.rt_self.__whitespace__(c) or self.rt_self.__punctuation__(c):
                if len(_word) > 0:
                    _txt.append(_word)
                    _typ.append('word')
                    _num.append(word_num)
                    _beg.append(i0)
                    _end.append(i)
                    word_num += 1
                    _word = ''
                    i0 = -1
            elif i0 != -1:
                _word += c
            else:
                i0    = i
                _word = str(c)
            i += 1
        if len(_word) > 0:
            _txt.append(_word)
            _typ.append('word')
            _num.append(word_num)
            _beg.append(i0)
            _end.append(i)
            word_num += 1
        
        return pd.DataFrame({
            'text': _txt,
            'type':_typ,
            'num':_num,
            'beg':_beg,
            'end':_end
        })
    
    #
    # renderDataFrame() - render a positional dataframe (assumes some level of filtering)
    # ... i.e., one can filter the pandas dataframe and then re-render to highlight text/etc.
    #
    def renderDataFrame(self, 
                        df,                               # Positional Dataframe from positionalDataFrame() method...
                        color_by          = None,         # Field in the dataframe
                        color_by_style    = 'highlight',  # 'highlight' (like a highlighter), 'underline', or 'text'
                        highlight_opacity = 0.6,          # Opacity of highlight
                        context_opacity   = 0.7,          # Context opacity
                        render_context    = True):        # Render all of the text as a background
        x,y,w,h = self.bounds
        _co     = self.rt_self.co_mgr.getTVColor('background','default')
        my_svg  = f'<svg x="0" y="0" width="{w}" height="{h}">' + \
                  f'<rect x="0" y="0" width="{w}" height="{h}" fill="{_co}" />'
        
        # Render the context for the highlights
        if render_context:
            my_svg += self.svg
            my_svg += f'<rect x="0" y="0" width="{w}" height="{h}" fill="{_co}" fill-opacity="{context_opacity}" />'
        
        # First pass... really only for highlights or for underlines
        _co = '#404040'
        if   color_by_style == 'highlight':
            for row_i,row in df.iterrows():
                if color_by is not None:
                    _co = self.rt_self.co_mgr.getColor(row[color_by])
                _poly = self.spanGeometry(row['beg'],row['end'])
                my_svg += f'<path d="{self.rt_self.shapelyPolygonToSVGPathDescription(_poly)}" fill="{_co}" fill-opacity="{highlight_opacity}" />'
                pass
        elif color_by_style == 'underline':
            _stroke_w = min(0.5 + self.txt_h/14, 2.5)
            for row_i,row in df.iterrows():
                if color_by is not None:
                    _co = self.rt_self.co_mgr.getColor(row[color_by])
                x0, x1, y = None,None,None
                for i in range(row['beg'],row['end']):
                    c, xy = self.txt[i], self.orig_to_xy[i]
                    if self.rt_self.__whitespace__(c) or self.rt_self.__punctuation__(c):
                        if x0 is not None:
                            my_svg += f'<line x1="{x0}" y1="{y + 1 + _stroke_w}" x2="{x1}" y2="{y + 1 + _stroke_w}" stroke="{_co}" stroke-width="{_stroke_w}" />'
                            x0, x1, y = None,None,None
                    elif x0 is not None and y != xy[1]:
                        my_svg += f'<line x1="{x0}" y1="{y + 1 + _stroke_w}" x2="{x1}" y2="{y + 1 + _stroke_w}" stroke="{_co}"  stroke-width="{_stroke_w}" />'
                        x0, x1, y = None,None,None
                    elif x0 is None:
                        x0, x1, y = xy[0], xy[0] + self.rt_self.textLength(c, self.txt_h), xy[1]
                    else:
                        x1 = xy[0] + self.rt_self.textLength(c, self.txt_h)
                if x0 is not None:
                    my_svg += f'<line x1="{x0}" y1="{y + 1 + _stroke_w}" x2="{x1}" y2="{y + 1 + _stroke_w}" stroke="{_co}"  stroke-width="{_stroke_w}" />'
        elif color_by_style == 'text':
            pass
        else:
            raise Exception('RTTextBlock.renderDataFrame() - color_by_style "{highlight_style}" unknown')

        # Second pass...
        for row_i, row in df.iterrows():
            for i in range(row['beg'],row['end']):
                xy = self.orig_to_xy[i]
                if color_by is None or color_by_style != 'text':
                    my_svg += self.rt_self.svgText(self.txt[i], xy[0], xy[1], self.txt_h)
                else:
                    _co = self.rt_self.co_mgr.getColor(row[color_by])
                    my_svg += self.rt_self.svgText(self.txt[i], xy[0], xy[1], self.txt_h, color=_co)

        my_svg += '</svg>'
        return my_svg